#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Odd_Ratio.py
# @Author  : Rensc


# from turtle import color
# from scipy.stats import fisher_exact
from joblib import Parallel, delayed
import pandas as pd
from statsmodels.sandbox.stats.multicomp import multipletests
import sys
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from . import RPFs


class OddRatio(object):
    def __init__(self, args):
        # input and output files
        self.rpf_file = args.rpf
        self.output_prefix = args.output

        # opts for odd ratio
        self.gene = args.list
        self.site = args.site
        self.frame = args.frame
        self.tis = args.tis
        self.tts = args.tts
        
        self.normal = args.normal
        self.flag = args.fdr
        self.pvalue = args.value
        self.fillna = args.zero

        self.thread = args.thread

        # samples
        self.control = args.control.split(',')
        self.treat = args.treat.split(',')
        self.rpf_num = args.min

        self.sample_name = self.control + self.treat
        self.sample_num = len(self.sample_name)

        # for the high expression gene filter
        self.merged_rpf = None
        self.raw_rpf = None
        self.high_gene = None
        self.high_rpf = None
        self.gene_rpf_sum = None
        self.total_rpf_num = None
        self.dim2_df = None
        self.merge_odd_ratio = None

        # summary stalling codon
        self.scale = args.scale
        self.codon_odd_ratio = None
        self.codon_odd_ratio_num = None
        self.codon_odd_ratio_prop = None

        # codon list
        self.stop = args.stop
        self.codon_dict, self.codon_table = RPFs.codon_table()

        if self.stop:
            del self.codon_dict['TAA']
            del self.codon_dict['TAG']
            del self.codon_dict['TGA']
            
            self.codon_table = self.codon_table[self.codon_table['Abbr'] != 'Stop']

    def read_rpf(self):
        '''
        @Message  : import the rpf density file and convert them to rpm density
        @Input    : self.rpf_file --> RPF density file
                    self.site --> the site of the codon
                    self.frame --> the frame of the codon
                    self.tis --> the start site of the CDS
                    self.tts --> the end site of the CDS
                    self.sample_num --> the number of the samples
                    self.sample_name --> the name of the samples
                    self.gene --> the gene list
                    self.rpf_num --> the minimal number of the RPFs
        @return   : merged_rpf, sample_name, sample_num, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        @Flow     : step 1. import the rpf density file with import_rpf function,
                    step 2. convert the rpf density to rpm density
                    step 3. drop the codon without RPFs coverage
                    step 4. fill the empty codon rpm value with minimal rpm value (>0),
        '''

        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf_file,
                                      sites=self.site, frame=self.frame,
                                      tis=self.tis, tts=self.tts,
                                      sample_num=self.sample_num, sample_name=self.sample_name,
                                      gene=self.gene, rpf_num=self.rpf_num)
        # self.raw_rpf = rpf_results[0].copy()
        self.sample_name = rpf_results[1]
        self.sample_num = rpf_results[2]
        self.merged_rpf = rpf_results[3].copy()
        self.total_rpf_num = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5]
        self.high_gene = rpf_results[6]
        # self.high_rpf = rpf_results[7].copy()
        del rpf_results
        
        self.merged_rpf = self.merged_rpf[self.merged_rpf['codon'].isin(self.codon_dict.keys())]

        # convert the rpf density to rpm density
        if self.normal:
            self.merged_rpf.loc[:, self.sample_name] = self.merged_rpf[self.sample_name] * 1e6 / self.total_rpf_num
            # change the float to int with ceil
            # self.merged_rpf.loc[:, self.sample_name] = self.merged_rpf[self.sample_name].apply(np.ceil)
            # self.merged_rpf.loc[:, self.sample_name] = self.merged_rpf[self.sample_name].astype(int)

        self.high_rpf = self.merged_rpf.loc[self.merged_rpf['name'].isin(self.high_gene)]
        
        # drop the codon without RPFs coverage
        print('Ignore the codon without RPFs coverage.', flush=True)
        self.high_rpf = self.high_rpf[self.high_rpf[self.sample_name].sum(axis=1) > 0]

        # fill emtpy rpf with 0.1 * minimal rpm value (> 0)
        if self.fillna:
            min_rpm = self.high_rpf[self.high_rpf[self.sample_name] !=0 ].min().min()
            self.high_rpf[self.sample_name] += min_rpm * 0.01
        else:
            self.high_rpf = self.high_rpf[self.high_rpf[self.sample_name].min(axis=1) > 0]

        # summary the gene rpm
        self.gene_rpf_sum = self.high_rpf.groupby('name')[self.sample_name].sum()

    def make_two_dimensional_table(self):
        '''
        @Message  : prepare two dimensional table for fisher exactly test.
        @Input    : self.high_rpf --> high expression rpf table
                    self.control --> control sample table
                    self.treat --> treat sample table
        @Return   : self.gene_rpf_sum --> sum of high expression gene table
        @Flow     : 1 --> sum the rpf count of each gene with control and treatment group
                    2 --> calculate the gene delta
                    delta = each gene sum - each codon of this gene
        '''

        self.high_rpf = self.high_rpf.reset_index(drop=True)

        if len(self.control) > 1:
            self.high_rpf.loc[:, 'control'] = np.array(self.high_rpf[self.control].sum(axis=1))
            self.gene_rpf_sum.loc[:, 'control_sum'] = np.array(self.gene_rpf_sum[self.control].sum(axis=1))
        else:
            self.high_rpf.loc[:, 'control'] = np.array(self.high_rpf[self.control])
            self.gene_rpf_sum.loc[:, 'control_sum'] = np.array(self.gene_rpf_sum[self.control])

        if len(self.treat) > 1:
            self.high_rpf.loc[:, 'treat'] = np.array(self.high_rpf[self.treat].sum(axis=1))
            self.gene_rpf_sum.loc[:, 'treat_sum'] = np.array(self.gene_rpf_sum[self.treat].sum(axis=1))
        else:
            self.high_rpf.loc[:, 'treat'] = np.array(self.high_rpf[self.treat])
            self.gene_rpf_sum.loc[:, 'treat_sum'] = np.array(self.gene_rpf_sum[self.treat])

        # rpf_sum = np.array([self.total_rpf_num[self.control].sum(), self.total_rpf_num[self.treat].sum()])
        # rpf_sum = np.array([self.high_rpf['control'].sum(), self.high_rpf['treat'].sum()])
        # self.high_rpf.loc[:, ['control_sum', 'treat_sum']] = rpf_sum - np.array(self.high_rpf[['control', 'treat']])

        self.gene_rpf_sum = self.gene_rpf_sum[['control_sum', 'treat_sum']].reset_index()
        self.high_rpf = pd.merge(self.high_rpf, self.gene_rpf_sum, on='name', how='left')

    def calc_odd_ratio_v1(self):
        '''
        @Message  : fisher exactly test with fisher package.
                    package fisher were developed with python version < 3.9
        @Input    : self.dim2_df --> two dimension table for fisher-exactly test
        @Return   : control --> rpm of control
                    treat --> rpm of treat
                    control_delta --> delta rpm of control, 
                    treat_delta --> delta rpm of treat samples
                    self.high_rpf --> contain the high rpf and odd-ratio
                    bhfdr --> calculate bhfdr with multipletests
        @Flow     : step1 --> run
        '''
        
        from fisher import pvalue_npy

        self.dim2_df = self.high_rpf[['control', 'treat', 'control_sum', 'treat_sum']]

        # set the table
        control = self.dim2_df['control'].to_numpy(dtype='float32')
        treat = self.dim2_df['treat'].to_numpy(dtype='float32')
        control_delta = self.dim2_df['control_sum'].to_numpy(dtype='float32')
        treat_delta = self.dim2_df['treat_sum'].to_numpy(dtype='float32')

        self.high_rpf.loc[:, ['control', 'treat']] = self.dim2_df.loc[:, ['control', 'treat']].copy()

        # calculate the odd ratio
        print('\n\n' + 'Ignore codon with the none RPFs coverage.', flush=True)
        np.seterr(divide='ignore', invalid='ignore')
        self.high_rpf.loc[:, 'odd2'] = (treat * control_delta) / (control * treat_delta)

        # calculate the pvalue
        p_left, p_right, p_two_tail = pvalue_npy(control, treat_delta, treat, control_delta)
        bhfdr = multipletests(p_two_tail, alpha=0.05, method='fdr_bh')

    def calc_odd_ratio(self):
        '''
        @Message  : calculate the odd ratio and pvalue with chi2-test
        @Input    : self.high_rpf --> high expression rpf table
        @Return   : self.high_rpf --> contain the high rpf and odd-ratio
        @Flow     : step1 --> calculate the delta of each gene
                    step2 --> calculate the odd ratio
        '''
        
        control = self.high_rpf['control'].to_numpy(dtype='float32')
        treat = self.high_rpf['treat'].to_numpy(dtype='float32')
        control_delta = self.high_rpf['control_sum'].to_numpy(dtype='float32') - control
        treat_delta = self.high_rpf['treat_sum'].to_numpy(dtype='float32') - treat
        
        np.seterr(divide='ignore', invalid='ignore')
        self.high_rpf.loc[:, 'odd'] = (treat / treat_delta) / (control / control_delta)

    def calc_chi2_test(self):
        '''
        @Message  : calculate the odd ratio and pvalue with chi2-test
        @Input    : self.high_rpf --> high expression rpf table
        @Return   : self.high_rpf --> contain the high rpf and odd-ratio and pvalue
        @Flow     : step1 --> calculate the odd ratio and pvalue
                    step2 --> merge all results into one table

        '''
        import statsmodels.api as sm

        # calculate the pvalue with chi2-test
        statistic_list = []
        pvalue_list = []

        for i in range(len(self.high_rpf)):
            if i % 5000 == 0:
                print('row number: ' + str(i), flush=True)

            n = self.high_rpf['control'][i]
            m = self.high_rpf['treat'][i]
            N = self.high_rpf['control_sum'][i]
            M = self.high_rpf['treat_sum'][i]

            # calculate the odd ratio with statmodels

            # from scipy.stats import fisher_exact
            # observed = np.array([[n, N - n], [m, M - m]])
            # statistic, p_val = fisher_exact(observed)

            # ATTENTION： fisher-exact test does not fit the float values !!!
            table = [[n, m], [N - n, M - m]]
            result = sm.stats.Table2x2(table)

            # table = sm.stats.Table([[n, m], [N - n, M - m]])
            # result = table.test_nominal_association()

            statistic_list.append(result.oddsratio)
            pvalue_list.append(result.oddsratio_pvalue())

        print('row number: ' + str(i), flush=True)

        # correct the pvalue with Benjamin-Hochberg
        bhfdr = multipletests(pvalue_list, alpha=0.05, method='fdr_bh')

        # merge all results
        self.high_rpf.loc[:, 'statistic'] = statistic_list
        self.high_rpf.loc[:, 'pvalue'] = pvalue_list
        self.high_rpf.loc[:, 'bhfdr'] = bhfdr[1]
        self.high_rpf.loc[:, 'flag'] = bhfdr[0]

    def calc_chi2_test2(self):
        '''
        @Message  : calculate the chi2-test odd ratio and pvalue with multiple progress
        @Input    : self.high_rpf --> high expression rpf table
        @Return   : self.high_rpf --> contain the high rpf and odd-ratio and pvalue
        @Flow     : step1 --> define the function to calculate the pvalue with multiple progress
                    step2 --> calculate the odd ratio and pvalue
                    step3 --> correct the pvalue with Benjamin-Hochberg
                    step4 --> merge all results into one table

        '''
        import statsmodels.api as sm

        # define the function to calculate the pvalue with multiple progress
        def calc_pvalue(n, m, N, M):
            table = [[n, m], [N - n, M - m]]
            result = sm.stats.Table2x2(table)
            return result.oddsratio_pvalue()

        # calculate the pvalue with multiple progress
        pvalue_list = Parallel(n_jobs=self.thread)(delayed(calc_pvalue)(n, m, N, M) for n, m, N, M in zip(self.high_rpf['control'],
                                                                                                    self.high_rpf['treat'],
                                                                                                    self.high_rpf['control_sum'],
                                                                                                    self.high_rpf['treat_sum']))
        
        # correct the pvalue with Benjamin-Hochberg
        bhfdr = multipletests(pvalue_list, alpha=0.05, method='fdr_bh')

        # merge all results
        self.high_rpf.loc[:, 'pvalue'] = pvalue_list
        self.high_rpf.loc[:, 'bhfdr'] = bhfdr[1]
        self.high_rpf.loc[:, 'flag'] = bhfdr[0]

    def output_odd_ratio(self):
        '''
        output the odd ratio table
        '''
        out_odd_ratio_txt = self.output_prefix + '_codon_odd_ratio.txt'
        self.merge_odd_ratio = self.high_rpf[(self.high_rpf['from_tis'] >= self.tis) &
                                             (self.high_rpf['from_tts'] <= -self.tts) &
                                             (self.high_rpf[self.flag] < self.pvalue)]
        self.merge_odd_ratio.to_csv(out_odd_ratio_txt, sep='\t', index=False)

    @staticmethod
    def scale_method(scale, oddratio):
        if scale == 'minmax':
            # relative_oddratio = (oddratio - oddratio.min()) / (oddratio.max() - oddratio.min())
            relative_oddratio = oddratio / oddratio.max()
        elif scale == 'zscore':
            relative_oddratio = (oddratio - oddratio.mean()) / oddratio.std()
        else:
            print('Unknown scale method.', flush=True)
            sys.exit()
        return relative_oddratio
    
    def summarize_odd_ratio(self):
        '''
        @Message  : summarize the odd ratio
        @Input    : self.high_rpf --> high expression rpf table
        @Return   : count_aa_odd --> the count of codon odd ratio
                    per_aa_odd --> the proportion of codon odd ratio
        @Flow     : step1 --> filter the significant codon odd ratio
                    step2 --> calculate the count of significant codon odd ratio
                    step3 --> calculate the proportion of significant codon odd ratio

        '''
        # output filename
        codon_odd_file = self.output_prefix + '_sum_codon_odd_ratio.txt'

        codon_odd_num_column = ['Codon', 'AA', 'Abbr', 
                                 'codon_sum', 'control_codon_sum', 'treat_codon_sum', 
                                 'control_mean_odd_ratio', 'treat_mean_odd_ratio',
                                 'control_number', 'treat_number']

        # filter the significant codon odd ratio
        codon_odd_sum_all = self.high_rpf.groupby('codon')['control'].count()
        codon_odd_sum_all_0 = self.high_rpf.loc[self.high_rpf['odd'] < 1].groupby('codon')['control'].count()
        codon_odd_sum_all_1 = self.high_rpf.loc[self.high_rpf['odd'] > 1].groupby('codon')['treat'].count()
        
        # summary the odd ratio of each codon
        codon_odd_mean_all_0 = self.high_rpf.loc[self.high_rpf['odd'] < 1].groupby('codon')['control'].mean()
        codon_odd_mean_all_1 = self.high_rpf.loc[self.high_rpf['odd'] > 1].groupby('codon')['treat'].mean()

        sig_high_rpf = self.high_rpf.loc[self.high_rpf[self.flag] < self.pvalue, :]
        codon_odd_num_0 = sig_high_rpf.loc[sig_high_rpf['odd'] < 1].groupby('codon')['control'].count()
        codon_odd_num_1 = sig_high_rpf.loc[sig_high_rpf['odd'] > 1].groupby('codon')['treat'].count()

        # sig oddratio number
        codon_odd_ratio = pd.concat([self.codon_table,
                                    codon_odd_sum_all, codon_odd_sum_all_0, codon_odd_sum_all_1, 
                                    codon_odd_mean_all_0, codon_odd_mean_all_1,
                                    codon_odd_num_0, codon_odd_num_1], 
                                    join='inner', axis=1)
        codon_odd_ratio.reset_index(inplace=True)
        codon_odd_ratio.columns = codon_odd_num_column

        # filter the stop codon
        if self.stop:
            codon_odd_ratio = codon_odd_ratio.loc[~codon_odd_ratio['Codon'].isin(["TGA", "TAG", "TAA"]), :]
            # codon_odd_ratio_prop = codon_odd_ratio_prop.loc[~codon_odd_ratio_prop['Codon'].isin(["TGA", "TAG", "TAA"]), :]
        else:
            pass

        # sig oddratio proportion
        codon_odd_ratio_prop = codon_odd_ratio.loc[:, ['control_number', 'treat_number']].div(codon_odd_ratio['codon_sum'], axis=0) * 100
        codon_odd_ratio_prop.columns = ['control_proportion', 'treat_proportion']
        
        # calculate the relative odd ratio
        codon_odd_ratio_num_rel = self.scale_method(self.scale, codon_odd_ratio.loc[:, ['control_number', 'treat_number']])
        codon_odd_ratio_prop_rel = self.scale_method(self.scale, codon_odd_ratio_prop)
        codon_odd_ratio_num_rel.columns = ['control_number_relative', 'treat_number_relative']
        codon_odd_ratio_prop_rel.columns = ['control_proportion_relative', 'treat_proportion_relative']

        self.codon_odd_ratio = pd.concat([codon_odd_ratio, codon_odd_ratio_prop, codon_odd_ratio_num_rel, codon_odd_ratio_prop_rel], axis=1)
        self.codon_odd_ratio.sort_values(by=['Abbr', 'Codon'], inplace=True)

        self.codon_odd_ratio['control_group'] = ",".join(self.control)
        self.codon_odd_ratio['treat_group'] = ",".join(self.treat)

        # melt the table
        self.codon_odd_ratio_num = pd.melt(self.codon_odd_ratio, 
                                    id_vars=['Codon', 'AA', 'Abbr'], value_vars=['control_number', 'treat_number'],
                                    var_name='groups', value_name='Number')
        self.codon_odd_ratio_num['Codon'] = self.codon_odd_ratio_num['Codon'] + '(' + self.codon_odd_ratio_num['Abbr'] + ')'

        self.codon_odd_ratio_prop = pd.melt(self.codon_odd_ratio, 
                                    id_vars=['Codon', 'AA', 'Abbr'], value_vars=['control_proportion', 'treat_proportion'],
                                    var_name='groups', value_name='Proportion')
        self.codon_odd_ratio_prop['Codon'] = self.codon_odd_ratio_prop['Codon'] + '(' + self.codon_odd_ratio_prop['Abbr'] + ')'

        # output the odd ratio table
        self.codon_odd_ratio.to_csv(codon_odd_file, sep='\t', index=False)


    def draw_odd_ratio_bar(self):
        '''
        @Message  : draw the odd ratio barplot
        @Input    : self.codon_odd_ratio_num --> the count of codon odd ratio
                    self.codon_odd_ratio_prop --> the proportion of codon odd ratio
        @Return   : odd_ratio_barplot.pdf

        '''

        out_pdf = self.output_prefix + "_odd_barplot.pdf"
        out_png = self.output_prefix + "_odd_barplot.png"
        # draw the odd ratio number of each codon
        matplotlib.use('AGG')

        fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(15, 12), dpi=300)
        sns.barplot(data=self.codon_odd_ratio_num, x='Codon', y='Number', hue='groups', ax=ax[0])
        sns.barplot(data=self.codon_odd_ratio_prop, x='Codon', y='Proportion', hue='groups', ax=ax[1])

        # add the gene annotation here, need to import the txt file
        ax[0].set_ylabel('Number')
        ax[1].set_ylabel('Proportion (%)')
        ax[1].set_xlabel('Codon')

        plt.sca(ax[0])
        plt.xticks(rotation=90)
        plt.sca(ax[1])
        plt.xticks(rotation=90)

        plt.suptitle('Significant differential pausing codon')
        plt.tight_layout()
        # plt.show()

        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)

    def draw_odd_ratio_line(self):
        '''
        @Message  : draw the odd ratio lineplot
        @Input    : self.codon_odd_ratio_num --> the count of codon odd ratio
                    self.codon_odd_ratio_prop --> the proportion of codon odd ratio
        @Return   : odd_ratio_lineplot.pdf

        '''
        out_pdf = self.output_prefix + "_odd_lineplot.pdf"
        out_png = self.output_prefix + "_odd_lineplot.png"

        # draw the odd ratio number of each codon
        matplotlib.use('AGG')

        fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(10, 6), dpi=300)

        flag = 1
        for aa in self.codon_odd_ratio_num['AA'].unique().tolist():
            tmp = self.codon_odd_ratio_num.loc[self.codon_odd_ratio_num['AA']==aa, :]
            if flag == 1:
                sns.lineplot(data=tmp, x='Codon', y='Number', linewidth=1.5, hue='groups', style='groups',
                              markers=['d','o'], legend = True, ax=ax[0])
                flag += 1
            else:
                sns.lineplot(data=tmp, x='Codon', y='Number', linewidth=1.5, hue='groups', style='groups',
                              markers=['d', 'o'], legend = False, ax=ax[0])

        flag = 1
        for aa in self.codon_odd_ratio_prop['AA'].unique().tolist():
            tmp = self.codon_odd_ratio_prop.loc[self.codon_odd_ratio_prop['AA']==aa, :]
            if flag == 1:
                sns.lineplot(data=tmp, x='Codon', y='Proportion', linewidth=1.5, hue='groups', style='groups',
                             markers=['d', 'o'], legend = True, ax=ax[1])
                flag += 1
            else:
                sns.lineplot(data=tmp, x='Codon', y='Proportion', linewidth=1.5, hue='groups', style='groups',
                             markers=['d', 'o'], legend = False, ax=ax[1])

        # add the gene annotation here, need to import the txt file
        ax[0].legend(loc=0, ncol=2, fontsize=6)
        ax[1].legend(loc=0, ncol=2, fontsize=6)

        ax[0].set_ylabel('Number')
        ax[1].set_ylabel('Proportion (%)')
        ax[1].set_xlabel('Codon')

        plt.sca(ax[0])
        plt.xticks(rotation=90)
        plt.sca(ax[1])
        plt.xticks(rotation=90)

        plt.suptitle('Number of sig. diff. pausing codon')
        plt.tight_layout()
        # plt.show()

        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)

    def draw_odd_ratio_scatter(self):
        '''
        @Message  : draw the odd ratio scatterplot
        @Input    : self.codon_odd_ratio_num --> the count of codon odd ratio
                    self.codon_odd_ratio_prop --> the proportion of codon odd ratio
        @Return   : odd_ratio_scatterplot.pdf

        '''

        out_pdf = self.output_prefix + "_odd_scatter.pdf"
        out_png = self.output_prefix + "_odd_scatter.png"

        # count_df = self.codon_odd_ratio_num.pivot_table(index=['Codon', 'AA', 'Abbr'], columns='groups', values='Number').reset_index()
        # per_df = self.codon_odd_ratio_prop.pivot_table(index=['Codon', 'AA', 'Abbr'], columns='groups', values='Proportion').reset_index()

        # draw the odd ratio number of each codon
        matplotlib.use('AGG')

        fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(9, 6.5), dpi=300)
        
        count_plot = sns.scatterplot(data=self.codon_odd_ratio, x='control_number', y='treat_number', 
                                     s = 80, hue='AA', legend = 'brief', ax=ax[0])
        count_plot.legend(loc=8, bbox_to_anchor=(0.5, -0.5), ncol=7, fontsize=6)

        per_plot = sns.scatterplot(data=self.codon_odd_ratio, x='control_proportion', y='treat_proportion', 
                                   s = 80, hue='AA', legend = 'brief', ax=ax[1])
        per_plot.legend(loc=8, bbox_to_anchor=(0.5, -0.5), ncol=7, fontsize=6)

        # add the gene annotation here, need to import the txt file
        ax[0].set_title('Number')
        ax[1].set_title('Proportion (%)')

        plt.suptitle('Number of sig. diff. pausing codon')
        plt.tight_layout()
        # plt.show()

        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : riboParser
@Script       : fq_split.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2024/05/14 11:07:12
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


import argparse
import gc
import time
import sys
import gzip
from argparse import RawTextHelpFormatter


# get arguments and define the scripts usage
def parse_args():
    parser = argparse.ArgumentParser(
        description="This script is used to split the fastq file with UMI-adapter file.",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s 1.0"
    )
    parser.add_argument(
        "-i", metavar="input", required=True, type=str, help="input fastq file"
    )
    parser.add_argument(
        "-a", metavar="adapter", required=True, type=str, help="input adapter table file, ID for output fastq file, Adapter for retrieve."
    )
    parser.add_argument(
        "-o", metavar="output", required=True, type=str, help="prefix of output file"
    )
    # adapter like this
    
    args = parser.parse_args()
    return args


# output splitd sequences
def fqsplit(fq_input, fq_output):
    r1_name = fq_output + ".R1.fq"

    with open(r1_name, 'w') as r1_out:
        with open(r2_name, 'w') as r2_out:

                for line in fq_input:
                    r1_out.writelines(line)
                    r1_out.writelines(fq_input.__next__())
                    r1_out.writelines(fq_input.__next__())
                    r1_out.writelines(fq_input.__next__())

                    r2_out.writelines(fq_input.__next__())
                    r2_out.writelines(fq_input.__next__())
                    r2_out.writelines(fq_input.__next__())
                    r2_out.writelines(fq_input.__next__())
                    
# gc collect
def clear():
    print("Clear Memory.")
    gc.collect()
    print("All done.\n")


# get time
def now_time():
    print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    sys.stdout.flush()


# main
def main():
    now_time()
    args = parse_args()
    print(args)
    if args.i.split('.')[-1] == 'gz':
        with gzip.open(args.i, 'rt') as fq_input:
            fqsplit(fq_input, args.o)
    else:
        with open(args.i, 'r') as fq_input:
            fqsplit(fq_input, args.o)
    clear()
    now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#@Project : riboParser
#@Script  : merge_rsem.py


import pandas as pd
import argparse

def rsem_merge_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to merge specified columns from result files')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        "-l", "--list", nargs='+', required=True, help="List for result files (e.g., '*results')."
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='output file name.'
    )

    # optional arguments
    parser.add_argument(
        "-c", "--column", required=False, choices=['expected_count', 'TPM', 'FPKM'], default='expected_count',
        help="Column name to merge (e.g., 'expected_count')."
    )

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def merge_files(file_list, column_name):
    '''
    @Message  : merge specified columns from result files.
    @Input    : list --> pattern for result files (e.g., '*results')
                column --> column name to merge (e.g., 'expected_count')
    @Return   : 
                output --> output file name for merged results
    @Flow     : step1 --> run
    '''
    
    merged_data = []
    for file in file_list:
        data = pd.read_csv(file, sep='\t', index_col=0)
        selected_column = data[[column_name]]
        # remove the path and suffix
        filename = file.split('/')[-1]
        selected_column.columns = [filename]
        merged_data.append(selected_column)

    return merged_data


def output_file(merged_data, output_file):
    
    merged_data = pd.concat(merged_data, axis=1)
    merged_data.to_csv(output_file, sep='\t')


def main():
    args = rsem_merge_args_parser()
    merged_data = merge_files(args.list, args.column)
    output_file(merged_data, args.output)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @time    : 2021/6/11 10:50
# @Project : riboParser
# @Script  : __init__.py.py
# @Version : python 3.8.5
# @Author  : Rensc
# @E-mail  : rensc0718@163.com



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Offset_SSCBM.py


import os.path
import sys
from collections import OrderedDict
from itertools import islice

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pysam
import seaborn as sns


class Mrna(object):

    def __init__(self, record):
        self.chromosome = record[0]
        self.gene_id = record[1]
        #self.gene_name = record[2]
        self.transcript_id = record[2]
        self.start = record[3]
        self.end = record[4]
        self.utr5_length = int(record[5])
        self.cds_length = int(record[6])
        self.utr3_length = int(record[7])
        self.strand = record[8]
        self.rep_transcript = record[9]
        self.modified = record[10]
        self.bam = []
        self.rpf = []
        self.seq = None


class Offset(object):

    def __init__(self, args):
        self.min_length = args.min
        self.max_length = args.max
        self.nt_num = self.max_length - self.min_length + 1
        self.mrna_file = args.transcript
        
        # set the offset to contain the raw reads offset, adj_tis_offset to alter the error offset.
        # self.column_name = ["length", "from_tis", "tis_5end", "tis_5end_per", "from_tts", "tts_3end", "tts_3end_per", "sum"]

        # self.tis_offset = {"start_codon": OrderedDict(), "stop_codon": OrderedDict()}
        self.tis_offset = {"tis_5end": OrderedDict(), "tis_3end": OrderedDict(),
                           "tts_5end": OrderedDict(), "tts_3end": OrderedDict()}
        self.adj_tis_offset = OrderedDict()
        self.tis_5end = None
        self.tis_3end = None
        self.tts_5end = None
        self.tts_3end = None
        self.detail = args.detail

        self.frame_offset = OrderedDict()
        self.merge_frame_offset = pd.DataFrame(columns=["length", "frame0", "rpfs0", "frame1", "rpfs1",
                                                        "frame2", "rpfs2", "p_site", "rpfs"])
        self.frame_offset_len = {}
        # self.mode = args.mode

        self.exp_offset = None
        self.shift_nt = args.shift

        self.mrna_dict = OrderedDict()
        self.length_dict = {}

        # arguments for bam file parsing
        self.silence = args.silence
        self.sample_file = args.bam
        self.sample_format = os.path.splitext(args.bam)[1]

        self.pysam_input = None
        self.pysam_output = None
        self.output_prefix = None

        self.bam_file_list = None
        self.sample_dict = None

        self.peak_length = args.exp_peak
        self.peak_reads = 0

        self.output_prefix = args.output

    def read_transcript(self):
        # read transcripts annotation, retrieve the TIS and TTS message for offset detection.
        trans_file = self.mrna_file

        with open(trans_file, 'r') as trans_file_in:
            for line in islice(trans_file_in, 1, None):
                record = line.strip().split('\t')
                now_mrna = Mrna(record)
                self.mrna_dict[now_mrna.transcript_id] = now_mrna

    def get_mrna_reads(self):

        # check the data type, to support the SAM/BAM file format both.
        if self.sample_format.lower() == ".bam":
            print("import file: {bam}.\n".format(bam=self.sample_file), flush=True)
            read_format = 'rb'
        elif self.sample_format.lower() == ".sam":
            print("import file: {bam}.\n".format(bam=self.sample_file), flush=True)
            read_format = 'r'
        else:
            print("Unknown file format, please input the correct bam or sam file.", flush=True)
            sys.exit()

        self.pysam_input = pysam.AlignmentFile(self.sample_file, read_format)

        # calculate the length distribution and eliminate the reads aligned to the negative strand.
        for line in self.pysam_input.fetch(until_eof=True):
            if line.reference_name in self.mrna_dict:
                read_length = line.infer_read_length()
                if line.is_reverse:
                    try:
                        self.length_dict[read_length][1] += 1
                    except KeyError:
                        self.length_dict[read_length] = [0, 1]
                else:
                    try:
                        self.length_dict[read_length][0] += 1
                    except KeyError:
                        self.length_dict[read_length] = [1, 0]
                    self.mrna_dict[line.reference_name].bam.append(line)
            elif self.silence:
                pass
            else:
                print(line.reference_name + ': not in reference file.', flush=True)

        peak_length, self.peak_reads = max(self.length_dict.items(), key=lambda length: length[1])
        
        # if self.peak_length:
        #     self.peak_length -= 1
        # else:
        #     self.peak_length = peak_length - 1

        self.pysam_input.close()

    def get_tis_offset(self):

        # build the offset table
        for number in range(self.min_length, self.max_length + 1):
            # self.tis_offset["tis_5end"][number] = OrderedDict({-i: 0 for i in range(self.max_length + 1, 0, -1)})
            self.tis_offset["tis_5end"][number] = OrderedDict({-i: 0 for i in range(self.max_length, -1, -1)})
            self.tis_offset["tis_3end"][number] = OrderedDict({i: 0 for i in range(self.max_length + 1)})
            # self.tis_offset["tts_5end"][number] = OrderedDict({-i: 0 for i in range(self.max_length + 1, 0, -1)})
            self.tis_offset["tts_5end"][number] = OrderedDict({-i: 0 for i in range(self.max_length, -1, -1)})
            self.tis_offset["tts_3end"][number] = OrderedDict({i: 0 for i in range(self.max_length + 1)})

        for mrna, mrna_attr in self.mrna_dict.items():
            # eliminate the transcript without read mapped
            if len(mrna_attr.bam) == 0:
                continue
            else:
                # get the TIS and TTS from mRNA dict.
                # 6:shift 3 nt for stop codon, shift 2 nt for p-site， and shift 1 for the python index start
                # cds_start, cds_end = mrna_attr.utr5_length, mrna_attr.utr5_length + mrna_attr.cds_length - 6
                cds_start, cds_end = mrna_attr.utr5_length, mrna_attr.utr5_length + mrna_attr.cds_length - 6

                for line in mrna_attr.bam:
                    map_start, map_end = line.get_blocks()[0]
                    # map start is not included, so add 1 nt.
                    # map_start = map_start + 1
                    map_start = map_start
                    read_length = line.infer_read_length()

                    # reads that match a specific length are retained.
                    # an other way is to calculate whole rpfs, and filter the min-max length in the plot module
                    if self.min_length <= read_length <= self.max_length:
                        # filter the possible range of offset lengths around the start codon.
                        if map_start <= cds_start <= map_end:
                            offset_5end = map_start - cds_start
                            offset_3end = map_end - cds_start
                            try:
                                self.tis_offset["tis_5end"][read_length][offset_5end] += 1
                            except KeyError:
                                # self.tis_offset["tis_5end"][read_length][offset_5end] = 1
                                pass
                            try:
                                self.tis_offset["tis_3end"][read_length][offset_3end] += 1
                            except KeyError:
                                pass
                                # self.tis_offset["tis_3end"][read_length][offset_3end] = 1

                        # filter the possible range of offset lengths around the stop codon.
                        elif map_start <= cds_end <= map_end:
                            offset_5end = map_start - cds_end
                            offset_3end = map_end - cds_end
                            try:
                                self.tis_offset["tts_5end"][read_length][offset_5end] += 1
                            except KeyError:
                                pass
                                # self.tis_offset["tts_5end"][read_length][offset_5end] = 1
                            try:
                                self.tis_offset["tts_3end"][read_length][offset_3end] += 1
                            except KeyError:
                                pass
                                # self.tis_offset["tts_3end"][read_length][offset_3end] = 1
                    else:
                        pass

        self.tis_5end = pd.DataFrame(self.tis_offset['tis_5end']).T
        self.tis_3end = pd.DataFrame(self.tis_offset['tis_3end']).T
        self.tts_5end = pd.DataFrame(self.tis_offset['tts_5end']).T
        self.tts_3end = pd.DataFrame(self.tis_offset['tts_3end']).T

        if self.detail:
            self.tis_5end.to_csv(self.output_prefix + '_' + 'tis_5end.txt', sep='\t', header=True, index=True)
            self.tis_3end.to_csv(self.output_prefix + '_' + 'tis_3end.txt', sep='\t', header=True, index=True)
            self.tts_5end.to_csv(self.output_prefix + '_' + 'tts_5end.txt', sep='\t', header=True, index=True)
            self.tts_3end.to_csv(self.output_prefix + '_' + 'tts_3end.txt', sep='\t', header=True, index=True)


    @staticmethod
    def offset_scale(offset):
        offset_norm = offset.sub(offset.mean(axis=1), axis=0)
        offset_norm_row = offset_norm.div(offset_norm.std(axis=1), axis=0)
        offset_norm_row.fillna(0, inplace=True)
        return offset_norm_row

    def shift_codon(self, now_psite, length):

        # shift the codon with 3nt to the normal p-site range
        if now_psite in range(7, 20):
            return now_psite

        if self.adj_tis_offset:
            if self.adj_tis_offset[length-1][-1]:
                if self.adj_tis_offset[length-1][-1] + 2 < now_psite:
                    return self.shift_codon(now_psite - 3, length)
                elif now_psite < self.adj_tis_offset[length-1][-1] - 2:
                    return self.shift_codon(now_psite + 3, length)
            else:
                return np.nan

        elif not self.adj_tis_offset:
            if now_psite > 19:
                return self.shift_codon(now_psite - 3, length)
            elif now_psite < 7:
                return self.shift_codon(now_psite + 3, length)

    def adjust_tis_offset(self):
        '''
        @Message  : function for .
        @Input    : param --> description
        @Return   : output --> description
        @Flow     : step1 --> merge the tis_5end and tts_5end to get the offset rpfs
                    step2 --> shift the offset to the normal p-site range
        '''
        
        # monosome data only needs to align p-site on one side.
        tis_5end = self.tis_5end.T.copy()
        tis_5end.index = abs(tis_5end.index)
        tts_5end = self.tts_5end.T.copy()
        tts_5end.index = abs(tts_5end.index)
        offset_rpfs = tis_5end + tts_5end

        for length in range(self.min_length, self.max_length + 1):
            # shift 1 nt for the both 5/3 reads end (5end + 3end = 2)
            # Empirical value: 2nt for Eukaryotes, 1nt for prokaryotes
            shift_nt = (length - self.peak_length) // self.shift_nt
            candidate_offset = [i for i in range(8 + shift_nt, 16 + shift_nt + 1)]

            # compare the max rpfs at same offset of 5end and 3end
            max_offset_site = offset_rpfs[length].idxmax()
            temp_offset_rpfs = offset_rpfs[length].copy()
            
            if temp_offset_rpfs[max_offset_site] == 0:
                max_offset_site = candidate_offset[5]
                
            while max_offset_site not in candidate_offset:
                del temp_offset_rpfs[max_offset_site]
                max_offset_site = temp_offset_rpfs.idxmax()

            # sum the max frame rpfs of [E-site, P-site, A-site], and the psite is the max_offset_site
            offset_frame0_range = range(max_offset_site - 3, max_offset_site + 6, 3)
            offset_frame1_range = range(max_offset_site - 2, max_offset_site + 6, 3)
            offset_frame2_range = range(max_offset_site - 1, max_offset_site + 6, 3)

            frame0_offset_rpfs = offset_rpfs[length][offset_frame0_range].sum()
            frame1_offset_rpfs = offset_rpfs[length][offset_frame1_range].sum()
            frame2_offset_rpfs = offset_rpfs[length][offset_frame2_range].sum()

            max_frame_offset_rpfs = max(frame0_offset_rpfs, frame1_offset_rpfs, frame2_offset_rpfs)

            # sum the all frame rpfs of [E-site, P-site, A-site]
            offset_sum = frame0_offset_rpfs + frame1_offset_rpfs + frame2_offset_rpfs
            max_offset_per = np.true_divide(max_frame_offset_rpfs, offset_sum)

            # psite = offset + 1
            psite = max_offset_site + 1
            psite = self.shift_codon(psite, length)

            self.adj_tis_offset[length] = [length,
                                           max_offset_site, frame0_offset_rpfs,
                                           max_offset_site + 1, frame1_offset_rpfs,
                                           max_offset_site + 2, frame2_offset_rpfs,
                                           psite, offset_sum, max_offset_per]


    def write_tis_offset(self):

        column_name = ["length", "frame0", "rpfs0", "frame1", "rpfs1", "frame2", "rpfs2", "p_site", "rpfs", "periodicity"]

        adj_tis_offset = pd.DataFrame(self.adj_tis_offset).T.copy()

        adj_tis_offset.columns = column_name
        adj_tis_offset["periodicity"] = adj_tis_offset["periodicity"] * 100
        adj_tis_offset["periodicity"] = adj_tis_offset["periodicity"].astype(float).round(2)

        adj_tis_offset[["length", "frame0", "rpfs0", "frame1", "rpfs1", "frame2", "rpfs2", "p_site",
                        "rpfs"]] = adj_tis_offset[["length", "frame0", "rpfs0", "frame1", "rpfs1", "frame2", "rpfs2", "p_site", "rpfs"]].astype(int)

        adj_tis_offset.sort_values(['length'], inplace=True)
        adj_tis_offset["ribo"] = ["first"] * self.nt_num
        adj_tis_offset.to_csv(self.output_prefix + "_sscbm_offset.txt", sep='\t', index=False)

    def draw_tis_heatmap(self):
        # draw the offset heatmap
        def draw_figure(tis_5end, tis_3end, tts_5end, tts_3end, out_pdf, out_png):
            matplotlib.use('Agg')
            now_cmap = 'Blues'

            fig = plt.figure(figsize=(12, 8), dpi=300)
            # start codon 5 end
            ax1 = plt.subplot(2, 2, 1)
            sns.heatmap(data=tis_5end, annot=None, linewidths=0.5, ax=ax1, cmap=now_cmap)
            # label_y = ax1.get_yticklabels()
            # plt.setp(label_y, rotation=0, horizontalalignment='right')
            ax1.set_title('RPFs 5end')
            ax1.set_ylabel("RPFs length")
            ax1.set_xlabel("from start codon")
            # start codon 3 end
            ax2 = plt.subplot(2, 2, 2)
            sns.heatmap(data=tis_3end, annot=None, linewidths=0.5, ax=ax2, cmap=now_cmap)
            # label_y = ax1.get_yticklabels()
            # plt.setp(label_y, rotation=0, horizontalalignment='right')
            ax2.set_title('RPFs 3end')
            ax2.set_ylabel("RPFs length")
            ax2.set_xlabel("from start codon")
            # stop codon 5 end
            ax3 = plt.subplot(2, 2, 3)
            sns.heatmap(data=tts_5end, annot=None, linewidths=0.5, ax=ax3, cmap=now_cmap)
            # label_y = ax1.get_yticklabels()
            # plt.setp(label_y, rotation=0, horizontalalignment='right')
            ax3.set_title('RPFs 5end')
            ax3.set_ylabel("RPFs length")
            ax3.set_xlabel("from stop codon")
            # stop codon 3 end
            ax4 = plt.subplot(2, 2, 4)
            sns.heatmap(data=tts_3end, annot=None, linewidths=0.5, ax=ax4, cmap=now_cmap)
            # label_y = ax1.get_yticklabels()
            # plt.setp(label_y, rotation=0, horizontalalignment='right')
            ax4.set_title('RPFs 3end')
            ax4.set_ylabel("RPFs length")
            ax4.set_xlabel("from stop codon")

            plt.tight_layout()
            # plt.show()
            fig.savefig(fname=out_pdf)
            fig.savefig(fname=out_png)

        # draw the raw offset heatmap
        out_pdf = self.output_prefix + "_sscbm_offset.pdf"
        out_png = self.output_prefix + "_sscbm_offset.png"
        draw_figure(self.tis_5end, self.tis_3end, self.tts_5end, self.tts_3end, out_pdf, out_png)

        # draw the scaled offset heatmap
        out_pdf_s = self.output_prefix + "_sscbm_offset_scale.pdf"
        out_png_s = self.output_prefix + "_sscbm_offset_scale.png"
        tis_5end_s = self.offset_scale(self.tis_5end)
        tis_3end_s = self.offset_scale(self.tis_3end)
        tts_5end_s = self.offset_scale(self.tts_5end)
        tts_3end_s = self.offset_scale(self.tts_3end)
        draw_figure(tis_5end_s, tis_3end_s, tts_5end_s, tts_3end_s, out_pdf_s, out_png_s)


    def make_frame_offset(self):
        '''
        make the offset dict

        1. peak of ribosome protect fragments about 30nt, 
        the read length distribution also varies significantly due to different protocols and species.
        So shift 1 nt (depend on the reads length) for the both 5/3 reads end to solve this problem (5end + 3end = 2).
        Empirical value: 2nt for Eukaryotes, 1nt for prokaryotes
        
        2. set the minimum and maximum window for p-site location
        3. set the first window for p-site calculation

        '''

        # assign the empirical value for the offset start
        self.exp_offset = 11 + self.peak_length - 30
        
        if self.adj_tis_offset:
            for length in range(self.min_length, self.max_length + 1):
                self.exp_offset = self.adj_tis_offset[length][1]

                left_offset1, left_offset2, left_offset3 = self.exp_offset, self.exp_offset + 1, self.exp_offset + 2

                self.frame_offset_len[length] = [left_offset1, left_offset2, left_offset3]
                self.frame_offset[length] = [0, 0, 0]
                
        else:
            for length in range(self.min_length, self.max_length + 1):
                shift_nt = (length - self.peak_length) // self.shift_nt
                if shift_nt < -3:
                    shift_nt = -3
                elif shift_nt > 6:
                    shift_nt = 6
                else:
                    pass

                left_offset1 = self.exp_offset + shift_nt
                left_offset2 = self.exp_offset + 1 + shift_nt
                left_offset3 = self.exp_offset + 2 + shift_nt

                self.frame_offset_len[length] = [left_offset1, left_offset2, left_offset3]
                self.frame_offset[length] = [0, 0, 0]

    def calc_frame(self, map_start, cds_start, reads_length, num1, num2, num3):
        '''
        calculate the offset frame of each reads

        1. trim the offset_length dependent on the reads length
        2. fit the offset length of each reads in the codon frame
        3. arrange the [0, 1, 2] frame reads into frame_offset dictionary
        '''

        if (map_start + self.frame_offset_len[reads_length][num1] - cds_start) % 3 == 0:
            try:
                self.frame_offset[reads_length][num1] += 1
            except KeyError:
                pass
        elif (map_start + self.frame_offset_len[reads_length][num2] - cds_start) % 3 == 0:
            try:
                self.frame_offset[reads_length][num2] += 1
            except KeyError:
                pass
        elif (map_start + self.frame_offset_len[reads_length][num3] - cds_start) % 3 == 0:
            try:
                self.frame_offset[reads_length][num3] += 1
            except KeyError:
                pass

    def get_mono_frame(self, mrna_attr, cds_start):
        '''
        through the entire bam file,
        1. reterieve the reads map start site of gene body
        2. Remove any reads that are too long or too short
        3. arrange predicted psite for each reads
        '''

        for line in mrna_attr.bam:
            map_start, map_end = line.get_blocks()[0]
            reads_length = line.infer_read_length()
            if self.min_length <= reads_length <= self.max_length:
                self.calc_frame(map_start, cds_start, reads_length, 0, 1, 2)

    def get_frame_offset(self):
        '''
        through the entire mrna dict file,
        1. define the frame offset dict file
        2. eliminate the transcript without read mapped
        3. set the gene body range for offset calculation, trim the stop codon from gene CDS
        4. check the category of sequencing profile
        monosome, disome, trisome
        5. disome and trisome need to perform two and three times psite calculations respectively

        '''

        self.make_frame_offset()

        for _, mrna_attr in self.mrna_dict.items():
            
            if len(mrna_attr.bam) == 0:
                continue
            else:
                cds_start, cds_end = mrna_attr.utr5_length, mrna_attr.utr5_length + mrna_attr.cds_length - 3 - 2
                self.get_mono_frame(mrna_attr, cds_start)


    def format_frame_offset(self):
        columns = ["length", "frame0", "rpfs0", "frame1", "rpfs1", "frame2", "rpfs2", "p_site", "rpfs"]

        # different flag for the disome and trisome sequencing profiles
        # next version
        ribo = [["first"], ["second"], ["third"]]
        frame_offset = pd.DataFrame(self.frame_offset).T
        rows, cols = frame_offset.shape[0], frame_offset.shape[1]
        
        def get_offset_len(temp_offset):
            for k, v in dict(temp_offset.idxmax(1)).items():
                if v == 0:
                    temp_offset.loc[k, "p_site"] = self.frame_offset_len[k][v] + 1
                elif v == 1:
                    temp_offset.loc[k, "p_site"] = self.frame_offset_len[k][v] + 1
                elif v == 2:
                    temp_offset.loc[k, "p_site"] = self.frame_offset_len[k][v] + 1 - 3

        # calculate the frame offset for each ribosome profile
        for part in range(0, cols, 3):
            temp_offset = frame_offset.iloc[:, part:part + 3].copy()
            get_offset_len(temp_offset)
            temp_offset["sum"] = temp_offset.apply(lambda x: x.iloc[0:3].sum(), axis=1)
            temp_offset.columns = ["rpfs0", "rpfs1", "rpfs2", "p_site", "rpfs"]
            temp_offset[["frame0", "frame1", "frame2"]] = [i[part:part + 3] for i in self.frame_offset_len.values()]
            temp_offset["length"] = temp_offset.index
            self.merge_frame_offset = pd.concat([self.merge_frame_offset, temp_offset[columns]])

        self.merge_frame_offset["periodicity"] = self.merge_frame_offset[["rpfs0", "rpfs1", "rpfs2"]].max(axis = 1) / self.merge_frame_offset["rpfs"]
        self.merge_frame_offset["periodicity"] = self.merge_frame_offset["periodicity"] * 100
        self.merge_frame_offset["periodicity"] = self.merge_frame_offset["periodicity"].astype(float).round(2)
        self.merge_frame_offset["ribo"] = ribo[part // 3] * rows
        self.merge_frame_offset[columns] = self.merge_frame_offset[columns].astype(int)


    def adjust_frame_offset(self):
        '''
        The length of reads varies in different sequencing files,
        so offset should be normalise to suitable length.

        1. first we need to check the peak length from offset dict

        2. In general, the offset codon frame of shorter reads will not be larger than the offset of peak reads length commonly,
        so the abnormal offset will shift to smaller offset in the range of one codon frame.

        3. Similarly, an offset codon frame of longer reads will not be shorter than peak reads length.

        '''

        # adjust the frame shift of reads shorter than peak length
        for length in range(self.peak_length, self.min_length, -1):
            psite1 = self.merge_frame_offset.loc[self.merge_frame_offset['length'] == length, 'p_site'].reset_index(drop=True)[0]
            psite2 = self.merge_frame_offset.loc[self.merge_frame_offset['length'] == length - 1, 'p_site'].reset_index(drop=True)[0]

            if psite2 + 2 < psite1:
                self.merge_frame_offset.loc[self.merge_frame_offset['length'] == length - 1, 'p_site'] = psite2 + 3
            elif psite2 - 2 > psite1:
                self.merge_frame_offset.loc[self.merge_frame_offset['length'] == length - 1, 'p_site'] = psite2 - 3
            else:
                continue

        # adjust the frame shift of reads longer than peak length
        for length in range(self.peak_length, self.max_length, 1):
            psite1 = self.merge_frame_offset.loc[self.merge_frame_offset['length'] == length, 'p_site'].reset_index(drop=True)[0]
            psite2 = self.merge_frame_offset.loc[self.merge_frame_offset['length'] == length + 1, 'p_site'].reset_index(drop=True)[0]
            
            if psite2 - 2 > psite1:
                self.merge_frame_offset.loc[self.merge_frame_offset['length'] == length + 1, 'p_site'] = psite2 - 3
            elif psite2 + 2 < psite1:
                self.merge_frame_offset.loc[self.merge_frame_offset['length'] == length + 1, 'p_site'] = psite2 + 3
            else:
                continue

    def write_frame_offset(self):
        '''
        arrange the frame offset with predicted psite
        '''
        for length in range(0, self.merge_frame_offset.shape[0]):
            offset_line = self.merge_frame_offset.iloc[length, :]
            offset_psite = offset_line.iloc[[1, 3, 5]]
            offset_rpfs = offset_line.iloc[[2, 4, 6]]
        
            offset_index = (offset_psite - offset_line.iloc[-4] + 1) % 3

            offset_psite.index = offset_index
            offset_psite = offset_psite.loc[[0, 1, 2]]

            offset_rpfs.index = offset_index
            offset_rpfs = offset_rpfs.loc[[0, 1, 2]]

            self.merge_frame_offset.iloc[length, [1, 3, 5]] = offset_psite
            self.merge_frame_offset.iloc[length, [2, 4, 6]] = offset_rpfs

        self.merge_frame_offset.to_csv(self.output_prefix + "_rsbm_offset.txt", sep='\t', index=False)

    def draw_frame_heatmap(self):
        out_pdf = self.output_prefix + "_rsbm_offset.pdf"
        out_png = self.output_prefix + "_rsbm_offset.png"

        # reterieve the psite and rpfs
        raw_frame_offset = self.merge_frame_offset.loc[:, ["length", "p_site", 'rpfs0', 'rpfs1', 'rpfs2']]

        # rename the ylabel
        offset_num = raw_frame_offset.shape[0]
        raw_frame_offset.index = [str(raw_frame_offset["length"].to_list()[i]) + '_' + str(raw_frame_offset["p_site"].to_list()[i]) for i in range(0, offset_num)]

        # format the offset table and rename the columns
        raw_frame_offset = raw_frame_offset.drop(columns=["length", "p_site"])
        raw_frame_offset.columns = ["frame0", "frame1", "frame2"]
        raw_frame_offset = raw_frame_offset.apply(pd.to_numeric)

        # scale the frame offset to range [0, 1]
        scale_frame_offset = self.offset_scale(raw_frame_offset)
        scale_frame_offset = scale_frame_offset.apply(pd.to_numeric)

        # draw the frame offset heatmap
        matplotlib.use('Agg')
        now_cmap = 'Blues'

        fig = plt.figure(figsize=(12, 6), dpi=300)

        # raw RPFs frame offset
        ax1 = plt.subplot(1, 2, 1)
        sns.heatmap(data=raw_frame_offset, annot=None, linewidths=0.5, ax=ax1, cmap=now_cmap)
        ax1.set_title('raw counts')
        ax1.set_ylabel("RPFs length")
        ax1.set_xlabel("open reading frame")

        # scale RPFs frame offset
        ax2 = plt.subplot(1, 2, 2)
        sns.heatmap(data=scale_frame_offset, annot=None, linewidths=0.5, ax=ax2, cmap=now_cmap)
        ax2.set_title('scaled counts')
        ax2.set_ylabel("RPFs length")
        ax2.set_xlabel("open reading frame")

        plt.tight_layout()
        # plt.show()
        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : riboParser
@Script       : revs.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2023/11/07 20:25:23
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


# import pandas as pd
# import polars as pl
# import numpy as np
# from collections import OrderedDict
# from Bio import SeqIO
import argparse

# get arguments and dim the scripts usage
def parse_args():
    parser = argparse.ArgumentParser(description="converted sequence into the reverse complementary strand")
    parser.add_argument('seq',type=str)
    arguments = parser.parse_args()
    return arguments

def revSeq(seq):
    revsd = ""
    for nt in seq.strip():
        if nt == "A":
            nt = "T"
            revsd =  nt + revsd
        elif nt == "a":
            nt = "t"
            revsd =  nt + revsd
        elif nt == "T":
            nt = "A"
            revsd =  nt + revsd
        elif nt == "t":
            nt = "a"
            revsd =  nt + revsd
        elif nt == "G":
            nt = "C"
            revsd =  nt + revsd
        elif nt == "g":
            nt = "c"
            revsd =  nt + revsd
        elif nt == "C":
            nt = "G"
            revsd =  nt + revsd
        elif nt == "c":
            nt = "g"
            revsd =  nt + revsd
        elif nt == "U":
            nt = "A"
            revsd =  nt + revsd
        elif nt == "u":
            nt = "a"
            revsd =  nt + revsd
        elif nt == "N":
            revsd =  nt + revsd
        else:
            print('illegal nulcleotide in your sequence: %s'%(nt))
            revsd = nt + revsd
    print('reversed  :\t%s'%seq[::-1])
    print('complement:\t%s'%revsd[::-1])
    print('rev + comp:\t%s'%(revsd))

def main():
    args = parse_args()
    revSeq(args.seq)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : riboParser
@Script       : fa_gc_sum.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2023/11/07 19:46:39
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


import pandas as pd
# import polars as pl
# import numpy as np
from collections import OrderedDict
from collections import Counter
# from Bio import SeqIO
import argparse


def fa_gc_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to calculate the GC content of fasta file.')

    require_group = parser.add_argument_group('Required arguments')
    require_group.add_argument(
        '-i', dest='input', required=True, type=str, 
        help='input fasta file name.'
    )
    require_group.add_argument(
        '-o', dest='output', required=True, type=str, 
        help='output txt file name.'
    )

    args = parser.parse_args()

    return args


def input_fa(args):
    '''
    @Message  : function for import the fasta file.
    @Input    : fasta --> input file
    @Return   : output --> fasta dict
    '''

    fa_dict = OrderedDict()
    seq_name = None

    with open(args.input, 'r') as fa_in:
        for line in fa_in:
            
            if line.startswith('>'):
                seq_name = line.strip()
                if seq_name in fa_dict:
                    fa_dict[seq_name + 'rep2'] = []
                else:
                    fa_dict[seq_name] = ['', 0, 0, 0, 0]
            else:
                fa_dict[seq_name][0] += line.strip()
    
    return fa_dict


def calc_gc(args, fa_dict):
    '''
    @Message  : function for calculate the GC content of fasta file.
    @Input    : fasta --> input file
    @Return   : output --> GC content of fasta file
    '''

    gc_dict = OrderedDict()

    for seq_name, seq_info in fa_dict.items():
        gc_dict[seq_name] = [seq_name,
                             seq_info[0].count('A'), 
                             seq_info[0].count('T'), 
                             seq_info[0].count('C'), 
                             seq_info[0].count('G')]

    gc_df = pd.DataFrame.from_dict(gc_dict, orient='index', columns=['name', 'A', 'T', 'C', 'G'])

    gc_df.to_csv(args.output, sep='\t', index=False)

    sum_gc = gc_df.iloc[:, 1:5].sum(axis=0)
    print(sum_gc)


def main():
    args = fa_gc_args_parser()
    fa_dict = input_fa(args)
    calc_gc(args, fa_dict)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : riboParser
@Script       : merge_bwt_log.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2024/01/16 15:47:02
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


import os
import pandas as pd
from collections import OrderedDict
import argparse

def stat_bwt_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to statstic the mapped reads from log files')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        "-l", "--list", nargs='+', required=True, help="List for bowtie mapping log files (e.g., '*log')."
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='prefix of output file name.'
    )

    # optional arguments
    parser.add_argument(
        "-n", "--name", required=False, default="rRNA,tRNA,ncRNA,mRNA,Genome", 
        help="set the name of each database (default: %(default)s)."
    )

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def process_log_files(log_file_list, database_list=None):
    '''
    @Message  : retrieve the mapped reads from log files.
    @Input    : list --> pattern for log files (e.g., '*log')
                name --> column name to merge (['rRNA', 'tRNA', 'ncRNA', 'mRNA', 'Genome'])
    @Return   : 
                output --> output nested dict contain the mapped reads
    @Flow     : step1 --> retrieve the mapped reads and unmapped reads
    '''
    
    if database_list is None:
        database_list = ['rRNA', 'tRNA', 'ncRNA', 'mRNA', 'Genome']
    else:
        database_list = database_list.split(',')

    result_dict = OrderedDict()

    # for each log file
    for log_file in log_file_list:
        file_prefix = os.path.basename(log_file).split('.')[0]
        result_dict[file_prefix] = OrderedDict()

        with open(log_file, 'r') as file_in:
            mapped_sum = 0
            flag = -1

            # for each line in log file
            for line in file_in:
                # retrieve the total reads
                if flag == -1 and line.startswith("# reads processed:"):
                    result_dict[file_prefix]['Total'] = int(line.split(' ')[-1])
                    flag += 1
                # retrieve the mapped reads
                elif flag >= 0 and line.startswith("Reported"):
                    result_dict[file_prefix][database_list[flag]] = int(line.split(' ')[-2])
                    mapped_sum += int(line.split(' ')[-2])
                    flag += 1
                # skip the useless lines
                elif flag >= 0 and line.startswith("# reads with") or line.startswith("# reads that") or line.startswith("# reads processed:"):
                    pass
                # retrieve the unmapped reads
                else:
                    print('Error: the log file is not correct.', flush=True)

            result_dict[file_prefix]['Others'] = result_dict[file_prefix]['Total'] - mapped_sum

    return result_dict, database_list


def output_table(result_dict, database_list, output_file):
    '''
    @Message  : function for output.
    @Input    : result_dict --> nested dict contain the mapped reads
    @Return   : output --> description
    @Flow     : step1 --> convert the dict to dataframe
    '''
    
    mapped_df = pd.DataFrame.from_dict(result_dict, orient='index').fillna(0)
    ratio_df = mapped_df.div(mapped_df['Total'], axis=0) * 100

    mapped_df = mapped_df.drop('Total', axis=1).reset_index().melt(id_vars='index', var_name='Database', value_name='Count')
    ratio_df = ratio_df.drop('Total', axis=1).reset_index().melt(id_vars='index', var_name='Database', value_name='Ratio')

    mapped_df = mapped_df.rename(columns={'index': 'Sample'}).set_index(['Sample', 'Database'])
    ratio_df = ratio_df.rename(columns={'index': 'Sample'}).set_index(['Sample', 'Database'])

    result_df = mapped_df.join(ratio_df, how='outer').reset_index()
    result_df['Ratio'] = result_df['Ratio'].astype('float').round(2)
    
    database_list = database_list + ['Others']
    result_df['Database'] = pd.Categorical(result_df['Database'], categories=database_list, ordered=True)
    result_df = result_df.sort_values(by=['Sample', 'Database'])

    result_df.to_csv(output_file + '_mapping.txt', sep='\t', index=False)


def main():

    print('Step1: Checking the input Arguments.', flush=True)
    args = stat_bwt_args_parser()

    print('Step2: improt the bowtie log file.', flush=True)
    result_dict, database_list = process_log_files(args.list, args.name)

    print('Step3: output the mapping table.', flush=True)
    output_table(result_dict, database_list, args.output)

    print('All done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_quant.py


from .ribo import ArgsParser
from .ribo import Quant


def main():

    ArgsParser.now_time()
    print('\nQuantify the RPFs in the different region.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.rpf_quant_args_parser()

    quant = Quant.Quant(args)

    print('\nStep2: Import the RPFs file.', flush=True)
    quant.read_rpf()

    print('\nStep3: Quantify the RPFs in different region.', flush=True)
    quant.quant_region()
    quant.output_total_rpf()

    print('\nStep4: Draw the RPFs bar plot of different region.', flush=True)
    quant.draw_rpf_barplot()

    print('\nStep5: Draw the RPFs cumulative plot of gene rpm.', flush=True)
    quant.draw_rpf_cdfplot()

    print('\nStep6: Draw the RPFs PCA plot of gene rpm.', flush=True)
    quant.draw_rpf_pcaplot()

    print('\nStep7: Draw the RPFs heatmap of gene rpm.', flush=True)
    quant.draw_rpf_heatmap2()

    print('\nAll done.', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @time    : 2021/8/28 23:00
# @Project : riboParser
# @Script  : retrieve_seq.py
# @Version : python 3.8.5
# @product : PyCharm
# @Author  : Rensc
# @E-mail  : rensc0718@163.com


import argparse
import time
import sys
from Bio import SeqIO


# get arguments and define the scripts usage
def parse_args():
    parser = argparse.ArgumentParser(description="This script is used to retrieve the fasta sequence by name.")
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s 1.0"
    )
    parser.add_argument(
        "-i", dest="input", required=True, type=str,
        help="input the fasta file"
    )
    parser.add_argument(
        "-n", dest="name", required=True, type=str,
        help="gene ids in txt format"
    )
    parser.add_argument(
        "-u", dest="unmapped", required=False, type=str,default='unmapped.ids',
        help="output the unmapped gene ids"
    )
    parser.add_argument(
        "-o", dest="output", required=True, type=str, default='results',
        help="prefix of output file name (default %(default)s_peaks.txt)"
    )
    args = parser.parse_args()

    args_dict = vars(args)
    for k, v in args_dict.items():
        print("%12s : %s" % (k, str(v)))

    sys.stdout.flush()
    return args


def import_gene_name(gene_name_file):
    gene_name_list = []
    with open(gene_name_file, 'r') as gene_name_in:
        for line in gene_name_in:
            gene_name_list.append(line.strip('>').strip())

    return gene_name_list


def retrieve_seq(gene_name_list, fasta_file, output_name, unmapped):
    out_fa = open(output_name, 'wb')
    out_ids = open(unmapped, 'w')
    fa_seq = SeqIO.index(fasta_file, "fasta")
    for gene in gene_name_list:
        try:
            out_fa.write(fa_seq.get_raw(gene))
        except KeyError:
            out_ids.write(str(gene) + '\n')


# main
def main():

    print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    sys.stdout.flush()
    args = parse_args()

    print('\nStep1: import the gene names.')
    gene_name_list = import_gene_name(args.name)

    print('\nStep2: retrieve the fasta sequences.')
    sys.stdout.flush()
    retrieve_seq(gene_name_list, args.input, args.output, args.unmapped)

    print("\nAll done!")
    print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))



if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_pausing.py


from .ribo import ArgsParser
from .ribo import Pausing


def main():
    ArgsParser.now_time()
    print('\nCalculate the relative codon pausing score.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.rpf_pausing_args_parser()
    pausing_score = Pausing.Pausing(args)

    print('\nStep2: Import the RPFs file.', flush=True)
    pausing_score.read_rpf()

    print('\nStep3: Calculate the ribosome pausing score.', flush=True)
    pausing_score.get_pausing_score()

    print('\nStep4: Output the ribosome pausing score.', flush=True)
    pausing_score.output_cds_pausing()
    pausing_score.output_cds_codon_pausing()
    pausing_score.output_sum_codon_pausing()
    pausing_score.output_all_pausing()

    print('\nStep5: Draw the codon pausing score.', flush=True)
    pausing_score.draw_codon_total_pausing_heat()
    pausing_score.draw_codon_valid_pausing_heat()
    # pausing_score.draw_codon_pausing_plot('total')
    # pausing_score.draw_codon_pausing_plot('valid')

    if args.figure != 'none':
        print('\nStep6: Draw the RPFs pausing score.', flush=True)
        pausing_score.draw_rpf_pausing_plot()
    else:
        print('\nStep6: Do not draw figures.', flush=True)

    print('\nAll done.', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : riboParser
@Script       : get_overlap_seq.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2023/03/10 10:45:16
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


import argparse
from Bio import SeqIO
from Bio import Seq

def args_parser():

    parser = argparse.ArgumentParser(description='This script is used to get overlap seq.')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        '-i', dest='input', required=True, type=str, help='input fasta file name.'
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='output file name.'
    )

    # optional arguments
    parser.add_argument(
        '-l', dest='length', required=False, type=int, default=6, help='nucleotide length (default: %(default)s nt).'
    )

    args = parser.parse_args()

    return args


def parse_fasta(fasta_file):
    # import the fasta file with biopython and return a dict
    sequences = {}

    record = SeqIO.parse(fasta_file, "fasta")
    for line in record:
        sequences[str(line.id)] = str(line.seq)

    return sequences


def search_overlap(sequences, nt_length):
    matched_sequences = []
    for id1, seq1 in sequences.items():
        for id2, seq2 in sequences.items():
            if id1 == id2:
                continue
            for i in range(nt_length, len(seq1) + 1):
                sub_seq = seq2[:i]
                if seq1.endswith(sub_seq) and len(sub_seq) > nt_length:
                    matched_sequences.append((id1, seq1, len(seq1), id2, seq2, len(seq2), sub_seq))
                    break

    return matched_sequences


def output_overlap_seq(matched_sequences, out_file):
    with open(out_file, 'w', newline='') as output:

        output.writelines('\t'.join(['id1', 'seq1', 'seq1_len', 'id2', 'seq2', 'seq2_len', 'overlap']) + '\n')
        for row in matched_sequences:
            now_row = [str(i) for i in row]
            output.writelines('\t'.join(list(now_row)) + '\n')


def main():
    args = args_parser()
    sequences = parse_fasta(args.input)
    matched_sequences = search_overlap(sequences, args.length)
    output_overlap_seq(matched_sequences, args.output)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : ArgsParser.py


import argparse
import os
import sys
import textwrap
import time


def now_time():
    print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()), flush=True)


def args_print(args):
    args_dict = vars(args)

    for k, v in args_dict.items():
        print("{:<12}:  {:<}".format(k, str(v)), flush=True)

    sys.stdout.flush()


def file_check(*files):
    for my_file in files:
        if os.path.exists(my_file):
            continue
        else:
            print("\nFile {file_name} is not exists!\n".format(file_name=my_file), flush=True)
            sys.exit()


def gtf_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to build the references.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-s', dest="sequence", required=True, type=str,
                             help="the input file name of genome sequence")
    input_group.add_argument('-t', dest="transcript", required=True, type=str, help="the input file name of gtf file")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (prefix + _norm.gtf)")

    # arguments for the modification
    parser.add_argument('-u', dest="utr", required=False, type=int, default=99,
                        help="add the pseudo UTR to the leaderless transcripts (default: %(default)s nt).")
    parser.add_argument('-c', dest="coding", required=False, action="store_true", default=False,
                        help="only retain the protein coding transcripts (default: %(default)s).")

    args = parser.parse_args()
    file_check(args.sequence, args.transcript)
    args_print(args)

    return args


def make_ribo_ref():
    parser = argparse.ArgumentParser(description="This script is used to build the references.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-g', dest="genome", required=True, type=str,
                             help="the input file name of genome sequence")
    input_group.add_argument('-t', dest="gtf", required=True, type=str, help="the input file name of gtf file")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (prefix + _norm.gtf)")

    # arguments for the modification
    parser.add_argument('-u', dest="utr", required=False, type=int, default=0,
                        help="add the pseudo UTR to the leaderless transcripts (default: %(default)s nt).")
    parser.add_argument('-c', dest="coding", required=False, action="store_true", default=False,
                        help="only retain the protein coding transcripts (default: %(default)s).")
    parser.add_argument('-l', dest="longest", required=False, action="store_true", default=False,
                        help="only retain the longest protein coding transcripts (default: %(default)s).")
    parser.add_argument('-w', dest="whole", required=False, action="store_true", default=False,
                        help="output whole message (default: %(default)s).")
    
    args = parser.parse_args()
    file_check(args.genome, args.gtf)
    args_print(args)

    return args


############################################################
def rpf_bam_check_parser():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="This script is used to summary the BAM condition.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-t', dest="transcript", required=True, type=str,
                             help="the input file name of gene annotation")
    input_group.add_argument('-b', dest="bam", required=True, type=str, help="the input file name of bam")
    input_group.add_argument('-o', dest="output", required=True, type=str, help="the prefix of output file.")

    # arguments for the modification
    parser.add_argument('--thread', dest="thread", required=False, type=int, default=1,
                        help='''the number of threads (default: %(default)s). 
                        Suitable for large bam files > 1G.
                        It will take a lot of memory.''')
    parser.add_argument('-g', dest="tag", choices=[0, 1], type=int, required=False, default=0,
                        help='''filter the number of reads mapped loci (default: %(default)s).
                        [0]: all reads will be used; [1]: reads with unique mapped loci will be used
                        ''')
    parser.add_argument('-a', dest="align", choices=['star', 'hisat2', 'bowtie2'], type=str, required=False,
                        default='star', help='''filter the number of reads mapped loci (default: %(default)s).
                        ''')
    parser.add_argument('-r', dest="reverse", action='store_true', required=False, default=False,
                        help="reads aligned to negative strand will also be counted. (default: %(default)s).")
    parser.add_argument('-l', dest="longest", action='store_true', required=False, default=False,
                        help="only keep the longest transcripts (default: %(default)s).")
    parser.add_argument('-s', dest="saturation", action='store_true', required=False, default=False,
                        help='''whether to calculate RPF saturation. (default: %(default)s).
                        This step will take a lot of time and memory.
                        ''')
    args = parser.parse_args()
    file_check(args.transcript, args.bam)
    args_print(args)

    return args


def digestion_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to Detect the digestion sites.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-t', dest="transcript", required=True, type=str,
                             help="the name of input transcript file in TXT format.")
    input_group.add_argument('-s', dest="sequence", required=True, type=str,
                             help="the name of input transcript sequence file in FA format.")
    input_group.add_argument('-b', dest="bam", required=True, type=str, help="the name of mapping file in BAM format.")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the name of output file. (prefix + _digestion_sites.txt)")

    # arguments for the ribo-seq parsing
    parser.add_argument('-l', dest="longest", action='store_true', required=False, default=False,
                        help="only retain the transcript with longest CDS of each gene (default: %(default)s)."
                             "Recommended : True")
    parser.add_argument('--scale', dest="scale", action='store_true', required=False, default=False,
                        help="only retain the transcript with longest CDS of each gene (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=20,
                        help="the minimum reads length to keep (default: %(default)s nt).")
    parser.add_argument('-M', dest="max", required=False, type=int, default=100,
                        help="the maximum reads length to keep (default: %(default)s nt).")

    args = parser.parse_args()
    file_check(args.transcript, args.sequence, args.bam)
    args_print(args)

    return args


def offset_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to detect the P-site offset.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-t', dest="transcript", required=True, type=str,
                             help="the name of input transcript filein TXT format.")
    input_group.add_argument('-b', dest="bam", required=True, type=str, help="the name of mapping file in BAM format.")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (prefix + _offset.txt)")

    # arguments for the offset detection
    parser.add_argument('--mode', dest="mode", required=False, type=str, default='SSCBM',
                        help="specify the mode of offset detect [SSCBM, RSBM]. (default: %(default)s).")
    parser.add_argument('-l', dest="longest", action='store_true', required=False, default=False,
                        help="only retain the transcript with longest CDS of each gene (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=27,
                        help="the minimum reads length to keep (default: %(default)s nt).")
    parser.add_argument('-M', dest="max", required=False, type=int, default=33,
                        help="the maximum reads length to keep (default: %(default)s nt).")
    parser.add_argument('-p', dest="exp_peak", required=False, type=int, default=30,
                        help="Expected RPFs peak length [~30 nt] (default: %(default)s nt).")
    parser.add_argument('-s', dest="shift", required=False, type=int, default=2,
                        help="psite shift for different RPFs length. (default: %(default)s nt).")
    parser.add_argument('--silence', dest="silence", required=False, action='store_true', default=True,
                        help="discard the warning information. (default: %(default)s).")
    parser.add_argument('-d', dest="detail", action='store_true', required=False, default=False,
                        help="output the details of offset (default: %(default)s).")
    args = parser.parse_args()
    file_check(args.transcript, args.bam)
    args_print(args)

    return args


def offset_rsbm_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to detect the P-site offset.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-t', dest="transcript", required=True, type=str,
                             help="the name of input transcript filein TXT format.")
    input_group.add_argument('-b', dest="bam", required=True, type=str, help="the name of mapping file in BAM format.")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (prefix + _offset.txt)")

    # arguments for the offset detection
    # parser.add_argument('--mode', dest="mode", required=False, type=str, default='SSCBM',
    #                     help="specify the mode of offset detect [RSBM, SSCBM]. (default: %(default)s).")
    parser.add_argument('-l', dest="longest", action='store_true', required=False, default=False,
                        help="only retain the transcript with longest CDS of each gene (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=27,
                        help="the minimum reads length to keep (default: %(default)s nt).")
    parser.add_argument('-M', dest="max", required=False, type=int, default=33,
                        help="the maximum reads length to keep (default: %(default)s nt).")
    parser.add_argument('-p', dest="exp_peak", required=False, type=int, default=29,
                        help="RPFs peak length [~30 nt] (default: %(default)s nt).")
    parser.add_argument('-e', dest="exp_offset", required=False, type=int, default=11,
                        help="expected offset length (default: %(default)s).")
    parser.add_argument('-s', dest="shift", required=False, type=int, default=2,
                        help="psite shift for different RPFs length."
                             "Empirical value: 2nt for Eukaryotes, 1nt for prokaryotes. (default: %(default)s nt).")
    parser.add_argument('--silence', dest="silence", required=False, action='store_true', default=True,
                        help="discard the warning information. (default: %(default)s).")
    parser.add_argument('-d', dest="detail", action='store_true', required=False, default=False,
                        help="output the details of offset (default: %(default)s).")
    args = parser.parse_args()
    file_check(args.transcript, args.bam)
    args_print(args)

    return args

def rna_offset_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to create the offset table of RNA-seq.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (prefix + _offset.txt)")

    # arguments for the offset table creation
    parser.add_argument('-m', dest="min", required=False, type=int, default=25,
                        help="the minimum reads length to keep (default: %(default)s nt).")
    parser.add_argument('-M', dest="max", required=False, type=int, default=151,
                        help="the maximum reads length to keep (default: %(default)s nt).")
    parser.add_argument('-e', dest="exp_offset", required=False, type=int, default=12,
                        help="Expected offset (default: %(default)s nt).")

    args = parser.parse_args()
    args_print(args)

    return args

def rpf_bam2bw_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to convert the bam to bedgraph.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-b', dest="bam", required=True, type=str,
                             help="name of input genome alignment file (bam/sam format).")
    input_group.add_argument('-p', dest="psite", required=True, type=str,
                            help="p-site offset file in TXT format. (default: %(default)s).")

    # arguments for the RPFs calculation
    parser.add_argument('-t', dest="times", required=False, type=int, default=3,
                        help="set the reads multiple aligned times. (default: %(default)s).")
    parser.add_argument('--second', dest="secondary", required=False, action='store_true', default=False,
                        help="discard the reads aligned secondary loci. (default: %(default)s).")
    parser.add_argument('--supply', dest="supplementary", required=False, action='store_true', default=False,
                        help="discard the supplementary reads (default: %(default)s ).")    
    parser.add_argument('-f', dest="format", required=False, choices=['bedgraph', 'wig'], type=str, default='bedgraph',
                        help="output the bed file. (default: %(default)s).")
    parser.add_argument('-n', dest="norm", required=False, action='store_true', default=False,
                        help="normalise the RPFs to RPM. (default: %(default)s).")
    parser.add_argument('-m', dest="merge", required=False, action='store_true', default=False,
                        help="merge minus and plus strand to one file. (default: %(default)s).")
    parser.add_argument('-o', dest="output", required=False, type=str,
                        help="output the bed file. (default: prefix + .bedgraph).")

    args = parser.parse_args()
    file_check(args.bam, args.psite)
    args_print(args)

    return args


def bam_filter_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to filter the reads length from bam file.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-i', dest="ibam", required=True, type=str,
                             help="the name of input bam file.")
    input_group.add_argument('-o', dest="obam", required=True, type=str,
                             help="the name of output bam file.")
    
    # arguments for the RPFs calculation
    parser.add_argument('-l', dest="length", required=False, type=str, default="27,28,29,30,31,32",
                        help="the minimum reads length to keep (default: %(default)s nt).")
    parser.add_argument('-u', dest="unique", required=False, action='store_true', default=False,
                        help="discard the secondary reads (default: %(default)s ).")
    parser.add_argument('-s', dest="supplementary", required=False, action='store_true', default=False,
                        help="discard the supplementary reads (default: %(default)s ).")
    parser.add_argument('-q', dest="quality", required=False, type=int, default=28,
                        help="discard the low mapping quality reads (default: %(default)s ).")

    args = parser.parse_args()
    file_check(args.ibam)
    args_print(args)

    return args


def ribo_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to convert Ribo-seq bam to p-site density.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-t', dest="transcript", required=True, type=str,
                             help="the name of input transcript file in TXT format.")
    input_group.add_argument('-s', dest="sequence", required=True, type=str,
                             help="the name of input transcript sequence file in FA format.")
    input_group.add_argument('-b', dest="bam", required=True, type=str, 
                             help="the name of mapping file in BAM format.")
    input_group.add_argument('-p', dest="psite", required=True, type=str,
                             help="the name of p-site offset file in TXT format.")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (output = prefix + _rpf.txt)")

    # arguments for the ribo-seq parsing
    parser.add_argument('-l', dest="longest", action='store_true', required=False, default=False,
                        help="only retain the transcript with longest CDS of each gene (default: %(default)s)."
                             "Recommended : True")
    parser.add_argument('-m', dest="min", required=False, type=int, default=27,
                        help="the minimum reads length to keep (default: %(default)s nt).")
    parser.add_argument('-M', dest="max", required=False, type=int, default=33,
                        help="the maximum reads length to keep (default: %(default)s nt).")
    input_group.add_argument('--period', dest='periodicity', required=False, type=float, default=40,
                             help="the minimum 3nt periodicity to keep. (default: %(default)s).")
    parser.add_argument('--silence', dest="silence", required=False, action='store_true', default=True,
                        help="discard the warning information. (default: %(default)s).")
    parser.add_argument('--thread', dest="thread", type=int, required=False, default=1,
                        help="the number of threads (default: %(default)s). It will take a lot of memory.")
    args = parser.parse_args()
    file_check(args.transcript, args.sequence, args.bam, args.psite)
    args_print(args)

    return args


def rna_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to convert RNA-seq bam to p-site density.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-t', dest="transcript", required=True, type=str,
                             help="the name of input transcript file in TXT format.")
    input_group.add_argument('-s', dest="sequence", required=True, type=str,
                             help="the name of input transcript sequence file in FA format.")
    input_group.add_argument('-b', dest="bam", required=True, type=str, help="the name of mapping file in BAM format.")
    input_group.add_argument('-p', dest="psite", required=True, type=str,
                             help="the name of p-site offset file in TXT format.")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (prefix + _rna.txt)")

    # arguments for the RNA-seq parsing
    parser.add_argument('-l', dest="longest", action='store_true', required=False, default=False,
                        help="only retain the transcript with longest CDS of each gene (default: %(default)s)."
                             " Recommended : True")
    parser.add_argument('-m', dest="min", required=False, type=int, default=25,
                        help="the minimum reads length to keep (default: %(default)s nt).")
    parser.add_argument('-M', dest="max", required=False, type=int, default=150,
                        help="the maximum reads length to keep (default: %(default)s nt).")
    parser.add_argument('-r', dest="rolling", action='store_true', required=False, default=False,
                        help="density is calculated once per ribosome width. (default: %(default)s)."
                             "only suitable for long RNA-seq reads.")
    parser.add_argument('--pe', dest="pair_end", action='store_true', required=False, default=False,
                        help="reads aligned to negative strand will also be counted. (default: %(default)s)."
                             "only suitable for pair-end RNA-seq reads.")
    parser.add_argument('--thread', dest="thread", type=int, required=False, default=1,
                        help="the number of threads (default: %(default)s). It will take a lot of memory.")
    args = parser.parse_args()
    file_check(args.transcript, args.sequence, args.bam, args.psite)
    args_print(args)

    return args


def shuffle_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to shuffle the RPFs data.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs density file in TXT format.")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (prefix + _shuffle.txt)")

    # arguments for the ribo-seq parsing
    parser.add_argument('-l', dest="list", required=False, type=str,
                        help="the list of input genes for transcript id.")
    parser.add_argument('-s', dest="seed", required=False, type=int, default=0,
                        help="the random seed for shuffle. (default: %(default)s).")
    parser.add_argument('-i', dest="individual", action='store_true', required=False, default=False,
                        help="shuffle the RPFs data for each samples. (default: %(default)s).")
    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def retrieve_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to retrieve density files.")

    # arguments for the required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs density file in TXT format.")
    input_group.add_argument('-o', dest="output", required=False, type=str, 
                             help="prefix of output file name (default: filename + '_retrieve.txt'.")

    # arguments for the ribo-seq parsing
    parser.add_argument('-l', dest="list", required=False, type=str,
                        help="the list of input genes for transcript id.")
    parser.add_argument('-m', dest="min", required=False, type=int, default=0,
                        help="retain transcript with more than minimum RPFs (default: %(default)s).")
    parser.add_argument('-n', dest="normal", action='store_true', required=False, default=False,
                        help="normalize the RPFs count to RPM (default: %(default)s).")
    parser.add_argument('-f', dest="format", action='store_true', required=False, default=False,
                        help="melt three column data of each sample to one column (default: %(default)s).")
    parser.add_argument('-s', dest="split", action='store_true', required=False, default=False,
                        help="split gene rpf to each TXT file (default: %(default)s).")

    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def rpf_merge_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to merge the density file.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-l', dest="list", required=True, type=str, help="the sample list in TXT format.")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (prefix + _rpf_merged.txt)")

    args = parser.parse_args()
    args_print(args)

    return args


def periodicity_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to draw the periodicity plot.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument(
        '-o', dest="output", required=True, type=str,
        help="the prefix of output file."
    )

    # arguments for the ribo-seq parsing
    input_group.add_argument('-t', dest="transcript", required=False, type=str,
                             help="the name of input transcript filein TXT format.")
    parser.add_argument('-m', dest="min", required=False, type=int, default=50,
                        help="retain transcript with more than minimum RPFs. (default: %(default)s).")
    parser.add_argument('--tis', dest="tis", required=False, type=int, default=0,
                        help="The number of codons after TIS will be discarded.. (default: %(default)s AA).")
    parser.add_argument('--tts', dest="tts", required=False, type=int, default=0,
                        help="The number of codons before TTS will be discarded.. (default: %(default)s AA).")
    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def metaplot_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to draw the meta plot.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-t', dest="transcript", required=True, type=str,
                             help="the name of input transcript filein TXT format.")
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix name of output file.")
    
    # arguments for the ribo-seq parsing
    parser.add_argument('-m', dest="min", required=False, type=int, default=50,
                        help="delete transcript with less than minimum RPFs. (default: %(default)s).")
    parser.add_argument('--utr5', dest="utr5", type=int, required=False, default=20,
                        help="the codon number in 5-utr region (default: %(default)s AA).")
    parser.add_argument('--cds', dest="cds", type=int, required=False, default=50,
                        help="the codon number in cds region (default: %(default)s AA).")
    parser.add_argument('--utr3', dest="utr3", type=int, required=False, default=20,
                        help="the codon number in 3-utr region (default: %(default)s AA).")
    parser.add_argument('-n', dest="normal", action='store_true', required=False, default=False,
                        help="normalize the RPFs count to RPM. (default: %(default)s).")
    parser.add_argument('--mode', dest="mode", required=False, choices=["line", "bar"], type=str, default='bar',
                        help="specify the mode of metaplot. (default: %(default)s).")
    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def coverage_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to draw the coverage meta plot.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-t', dest="transcript", required=True, type=str,
                             help="the name of input transcript filein TXT format.")
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('-o', dest="output", required=False, type=str, help="the prefix of output file.")

    # arguments for the ribo-seq parsing
    parser.add_argument('-f', dest="frame", choices=['0', '1', '2', 'all'], required=False, type=str, default='all',
                        help="set the reading frame for occupancy calculation. (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=50,
                        help="retain transcript with more than minimum RPFs. (default: %(default)s).")
    parser.add_argument('-b', dest="bin", required=False, type=str, default='30,100,30',
                        help="adjust the transcript to specified bins. 30 for 5'-UTR"
                             "and 3'-UTR, 100 for CDS. (default: %(default)s).")
    parser.add_argument('-n', dest="normal", action='store_true', required=False, default=False,
                        help="normalize the RPFs count to RPM. (default: %(default)s).")
    parser.add_argument('--thread', dest="thread", type=int, required=False, default=1,
                        help="the number of threads. (default: %(default)s).")
    parser.add_argument('--outlier', dest="outlier", action='store_true', required=False, default=False,
                        help="filter the outliers (default: %(default)s).")
    parser.add_argument('--set', dest="set", choices=['intersect', 'union'], required=False, type=str, default='union',
                        help="filter the gene list with 5-UTR / CDS / 3-UTR. (default: %(default)s).")
    parser.add_argument('--heat', dest="heatmap", action='store_true', required=False, default=False,
                        help="draw the coverage heatmap of whole gene (This step will take a long time). (default: %(default)s).")
    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def rpf_corr_args_parser():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="This script is used to draw the correlation of rpf density.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (prefix + _rpf_merged.txt)")

    args = parser.parse_args()
    args_print(args)

    return args


def meta_codon_plot_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to draw the meta codon plot.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-l', dest="list", required=False, type=str, default=None,
                             help="the gene name list in TXT format. (default: whole).")
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('-c', dest="codon", required=False, type=str, help="the codon list in TXT format.")
    input_group.add_argument('-o', dest="output", required=True, type=str, help="the prefix of output file.")

    # arguments for the ribo-seq parsing
    parser.add_argument('-f', dest="frame", choices=['0', '1', '2'], required=False, type=str, default='all',
                        help="set the reading frame for occupancy calculation. (default: %(default)s).")
    parser.add_argument('-a', dest="around", required=False, type=int, default=20,
                        help="retrieve length of codon upstream and downstream. (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=50,
                        help="retain transcript with more than minimum RPFs. (default: %(default)s).")
    parser.add_argument('--tis', dest="tis", required=False, type=int, default=0,
                        help="The number of codons after TIS will be discarded.. (default: %(default)s AA).")
    parser.add_argument('--tts', dest="tts", required=False, type=int, default=0,
                        help="The number of codons before TTS will be discarded.. (default: %(default)s AA).")
    parser.add_argument('-n', dest="normal", action='store_true', required=False, default=False,
                        help="normalize the RPFs count to RPM. (default: %(default)s).")
    parser.add_argument('-u', dest="unique", action='store_true', required=False, default=False,
                        help="delete the cross repetition codon in different window. (default: %(default)s).")
    parser.add_argument('-s', dest="scale", action='store_true', required=False, default=False,
                        help="scale the window density with gene density. (default: %(default)s).")
    parser.add_argument('--smooth', dest="smooth", required=False, default=None, type=str,
                        help="smooth the window density [eg, 3,1]. (default: %(default)s).")
    parser.add_argument('--thread', dest="thread", type=int, required=False, default=1,
                        help="the number of threads (default: %(default)s).")
    parser.add_argument('--fig', dest="fig", required=False, action='store_true', default=False,
                        help="output the figure. (default: %(default)s).")
    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def codon_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to draw the codon usage plot.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('-l', dest="list", required=False, type=str, default=None,
                             help="the gene name list in TXT format. (default: whole).")
    input_group.add_argument('-o', dest="output", required=True, type=str, help="the prefix of output file.")

    # arguments for the ribo-seq parsing
    parser.add_argument('-s', dest="site", choices=['E', 'P', 'A'], required=False, type=str, default='P',
                        help="set the E/P/A-site for occupancy calculation. (default: %(default)s).")
    parser.add_argument('-f', dest="frame", choices=['0', '1', '2', 'all'], required=False, type=str, default='all',
                        help="set the reading frame for occupancy calculation. (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=30,
                        help="retain transcript with more than minimum RPFs. (default: %(default)s).")
    parser.add_argument('--tis', dest="tis", required=False, type=int, default=15,
                        help="The number of codons after TIS will be discarded.. (default: %(default)s AA).")
    parser.add_argument('--tts', dest="tts", required=False, type=int, default=5,
                        help="The number of codons before TTS will be discarded.. (default: %(default)s AA).")
    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def rpf_gene_plot_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to draw the gene rpf plot.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    gene_name = parser.add_mutually_exclusive_group()
    gene_name.add_argument('-l', dest="list", required=False, type=str, help="the gene name list in TXT format.")
    gene_name.add_argument('-g', dest="gene", required=False, type=str, help="the gene name.")

    # arguments for the RPFs calculation
    parser.add_argument('--log', dest="log", required=False, choices=['2', '10'], type=str, default='Not',
                        help="set the y-axis to log scaling. (default: %(default)s).")

    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def rpf_pausing_args_parser():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="This script is used to calculate the relative pausing score.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('-l', dest="list", required=False, type=str, default=None,
                             help="the gene name list in TXT format. (default: whole).")
    input_group.add_argument('-o', dest="output", required=True, type=str, help="the prefix of output file.")

    # arguments for the pausing calculation
    parser.add_argument('-s', dest="site", choices=['E', 'P', 'A'], required=False, type=str, default='P',
                        help="set the E/P/A-site for pausing calculation. (default: %(default)s).")
    parser.add_argument('-f', dest="frame", choices=['0', '1', '2', 'all'], required=False, type=str, default='all',
                        help="set the reading frame for pausing calculation. (default: %(default)s).")
    parser.add_argument('-b', dest="background", required=False, type=int, default=2,
                        help="set the codon number before and after p-site as the background. (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=50,
                        help="retain transcript with more than minimum RPFs. (default: %(default)s).")
    parser.add_argument('--tis', dest="tis", required=False, type=int, default=10,
                        help="The number of codons after TIS will be discarded. (default: %(default)s AA).")
    parser.add_argument('--tts', dest="tts", required=False, type=int, default=5,
                        help="The number of codons before TTS will be discarded. (default: %(default)s AA).")
    parser.add_argument('-n', dest="normal", action='store_true', required=False, default=False,
                        help="normalize the RPFs count to RPM. (default: %(default)s).")
    parser.add_argument('--scale', dest="scale",  choices=['zscore', 'minmax'], required=False, type=str, default='minmax',
                        help="normalize the pausing score. (default: %(default)s).")
    parser.add_argument('--stop', dest="stop",  action='store_true', required=False, default=False,
                        help="rmove the stop codon. (default: %(default)s).")    
    parser.add_argument('--fig', dest="figure", choices=['none', 'png', 'pdf'], required=False, default='none',
                        help="draw the rpf pausing score of each gene (it will takes a lot of time). (default: %(default)s).")
    parser.add_argument('--all', dest="all", action='store_true', required=False, default=False,
                        help="output all pausing score of each gene. (default: %(default)s).")

    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def rpf_occupancy_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to draw the codon occupancy plot.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('-l', dest="list", required=False, type=str, default=None,
                             help="the gene name list in TXT format. (default: whole).")
    input_group.add_argument('-o', dest="output", required=True, type=str, help="the prefix of output file.")

    # arguments for the RPFs calculation
    parser.add_argument('-s', dest="site", choices=['E', 'P', 'A'], required=False, type=str, default='P',
                        help="set the E/P/A-site for occupancy calculation. (default: %(default)s).")
    parser.add_argument('-f', dest="frame", choices=['0', '1', '2', 'all'], required=False, type=str, default='all',
                        help="set the reading frame for occupancy calculation. (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=30,
                        help="retain transcript with more than minimum RPFs. (default: %(default)s).")
    parser.add_argument('-n', dest="normal", action='store_true', required=False, default=False,
                        help="normalize the RPFs count to RPM. (default: %(default)s).")
    parser.add_argument('--tis', dest="tis", required=False, type=int, default=15,
                        help="The number of codons after TIS will be discarded. (default: %(default)s AA).")
    parser.add_argument('--tts', dest="tts", required=False, type=int, default=5,
                        help="The number of codons before TTS will be discarded.. (default: %(default)s AA).")
    parser.add_argument('--scale', dest="scale",  choices=['zscore', 'minmax'], required=False, type=str, default='minmax',
                        help="normalize the occupancy. (default: %(default)s).")
    parser.add_argument('--stop', dest="stop",  action='store_true', required=False, default=False,
                        help="rmove the stop codon. (default: %(default)s).")
    parser.add_argument('--all', dest="all", action='store_true', required=False, default=False,
                        help="output all RPFs density. (default: %(default)s).")
    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def rpf_cdt_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to draw the codon decoding time plot.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('--rpf', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('--rna', dest="rna", required=True, type=str,
                             help="the name of input reads file in TXT format.")
    input_group.add_argument('-l', dest="list", required=True, type=str, default=None,
                             help="the gene name list in TXT format. (default: whole).")
    input_group.add_argument('-o', dest="output", required=True, type=str, help="the prefix of output file.")

    # arguments for the RPFs calculation
    parser.add_argument('-s', dest="site", choices=['E', 'P', 'A'], required=False, type=str, default='P',
                        help="set the E/P/A-site for codon decoding time calculation. (default: %(default)s).")
    parser.add_argument('-f', dest="frame", choices=['0', '1', '2', 'all'], required=False, type=str, default='all',
                        help="set the reading frame for codon decoding time calculation. (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=30,
                        help="retain transcript with more than minimum RPFs. (default: %(default)s).")
    # parser.add_argument('-n', dest="normal", action='store_true', required=False, default=False,
    #                     help="normalize the RPFs count to RPM. (default: %(default)s).")
    parser.add_argument('--tis', dest="tis", required=False, type=int, default=15,
                        help="The number of codons after TIS will be discarded.. (default: %(default)s AA).")
    parser.add_argument('--tts', dest="tts", required=False, type=int, default=5,
                        help="The number of codons before TTS will be discarded.. (default: %(default)s AA).")
    parser.add_argument('--scale', dest="scale",  choices=['zscore', 'minmax'], required=False, type=str, default='minmax',
                        help="normalize the codon decoding time. (default: %(default)s).")
    parser.add_argument('--stop', dest="stop",  action='store_true', required=False, default=False,
                        help="rmove the stop codon. (default: %(default)s).")
    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def rpf_cst_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to draw the codon decoding time plot.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('--rpf', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('--rna', dest="rna", required=True, type=str,
                             help="the name of input reads file in TXT format.")
    input_group.add_argument('-l', dest="list", required=False, type=str, default=None,
                             help="the gene name list in TXT format. (default: whole).")
    input_group.add_argument('-o', dest="output", required=True, type=str, help="the prefix of output file.")

    # arguments for the RPFs calculation
    parser.add_argument('-s', dest="site", choices=['E', 'P', 'A'], required=False, type=str, default='P',
                        help="set the E/P/A-site for codon decoding time calculation. (default: %(default)s).")
    parser.add_argument('-f', dest="frame", choices=['0', '1', '2', 'all'], required=False, type=str, default='all',
                        help="set the reading frame for codon decoding time calculation. (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=30,
                        help="retain transcript with more than minimum RPFs. (default: %(default)s).")
    parser.add_argument('-t', dest="times", required=False, type=int, default=10,
                        help="Specify the number of iteration times required for computation. (default: %(default)s).")
    parser.add_argument('--tis', dest="tis", required=False, type=int, default=0,
                        help="The number of codons after TIS will be discarded. (default: %(default)s AA).")
    parser.add_argument('--tts', dest="tts", required=False, type=int, default=0,
                        help="The number of codons before TTS will be discarded. (default: %(default)s AA).")
    parser.add_argument('--scale', dest="scale",  choices=['zscore', 'minmax'], required=False, type=str, default='minmax',
                        help="normalize the codon selection time. (default: %(default)s).")
    parser.add_argument('--stop', dest="stop",  action='store_true', required=False, default=False,
                        help="rmove the stop codon. (default: %(default)s).")
    args = parser.parse_args()
    file_check(args.rpf)
    file_check(args.rna)
    args_print(args)

    return args


def rpf_cov_args_parser():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="This script is used to calculate CoV in the CDS region.",
                                     epilog=textwrap.dedent('''\
        The group file needs to contain at least two column:
        +----------+---------+
        | name     | group  |
        +==========+=========+
        | wt1      | wt      |
        | wt2      | wt      |
        | treat1   | treat   |
        | treat2   | treat   |
        | ko1      | ko      |
        | ko2      | ko      |
        +----------+---------+
    '''))

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('-g', dest="group", required=False, type=str, default=None,
                            help="specify the list of sample group. (default: None)")
    input_group.add_argument('-l', dest="list", required=False, type=str, default=None,
                             help="the gene name list in TXT format. (default: whole).")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (prefix + _Cov.txt)")

    # arguments for the CoV calculation
    parser.add_argument('-f', dest="frame", choices=['0', '1', '2', 'all'], required=False, type=str, default='all',
                        help="set the reading frame for occupancy calculation. (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=5,
                        help="retain transcript with more than minimum RPFs. (default: %(default)s).")
    parser.add_argument('-n', dest="normal", action='store_true', required=False, default=False,
                        help="normalize the RPFs count to RPM. (default: %(default)s).")
    parser.add_argument('--tis', dest="tis", required=False, type=int, default=15,
                        help="The number of codons after TIS will be discarded. (default: %(default)s).")
    parser.add_argument('--tts', dest="tts", required=False, type=int, default=5,
                        help="The number of codons before TES will be discarded. (default: %(default)s).")
    parser.add_argument('--fig', dest="figure", action='store_true', required=False, default=False,
                        help="show the figure. (default: %(default)s).")
    
    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)
    return args


def cumulative_cov_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to calculate the cumulative CoV.")

    # arguments for the required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs density file in TXT format.")
    input_group.add_argument('-o', dest="output", required=False, type=str, 
                             help="prefix of output file name (default: filename + '_cumulative_CoV.txt'.")

    # arguments for the ribo-seq parsing
    parser.add_argument('-l', dest="list", required=False, type=str,
                        help="the list of input genes for transcript id.")
    parser.add_argument('-m', dest="min", required=False, type=int, default=0,
                        help="retain transcript with more than minimum RPFs (default: %(default)s).")
    parser.add_argument('-n', dest="normal", action='store_true', required=False, default=False,
                        help="normalize the RPFs count to RPM (default: %(default)s).")
    
    parser.add_argument('-t', dest="trim", required=False, type=int, default=50,
                        help="trim transcript with specific length (default: %(default)s nt).")
    
    parser.add_argument('-s', dest="split", action='store_true', required=False, default=False,
                        help="split gene rpf to each TXT file (default: %(default)s).")
    parser.add_argument('-z', dest="zero", action='store_true', required=False, default=False,
                        help="set the start site to zero (default: %(default)s).")
    
    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def rpf_odd_ratio_args_parser():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="This script is used to calculate the odd ratio of stalling codon.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('-l', dest="list", required=False, type=str, default=None,
                             help="the gene name list in TXT format. (default: whole).")
    input_group.add_argument('-o', dest="output", required=True, type=str, help="the prefix of output file.")

    # arguments for the odd ratio calculation
    parser.add_argument('--thread', dest="thread", required=False, type=int, default=1,
                        help="set the thread number. (default: %(default)s).")
    
    parser.add_argument('-s', dest="site", choices=['E', 'P', 'A'], required=False, type=str, default='P',
                        help="set the E/P/A-site for odd ratio calculation. (default: %(default)s).")
    parser.add_argument('-f', dest="frame", choices=['0', '1', '2', 'all'], required=False, type=str, default='all',
                        help="set the reading frame for odd ratio calculation. (default: %(default)s).")
    parser.add_argument('-c', dest="control", required=True, type=str,
                        help="specify the name of control samples, separated by commas.")
    parser.add_argument('-t', dest="treat", required=True, type=str,
                        help="specify the name of treat samples, separated by commas.")
    
    parser.add_argument('-n', dest="normal", action='store_true', required=False, default=False,
                        help="normalize the RPFs count to RPM. (default: %(default)s).")
    parser.add_argument('-z', dest="zero", action='store_true', required=False, default=False,
                        help="fill the empty codon with [0.01 * minimal value]. (default: %(default)s).")
    parser.add_argument('-m', dest="min", required=False, type=int, default=50,
                        help="specify the minimum count of RPFs per transcript to keep. (default: %(default)s).")
    parser.add_argument('--fdr', dest="fdr", required=False, type=str, choices=['bhfdr', 'pvalue'], default='bhfdr',
                        help="specify the multipletests. (default: %(default)s).")
    parser.add_argument('-v', dest="value", required=False, type=float, default=0.05,
                        help="specify the p value. (default: %(default)s).")
    
    parser.add_argument('--tis', dest="tis", required=False, type=int, default=0,
                        help="specify the number of codons after TIS will be discarded.. (default: %(default)s AA).")
    parser.add_argument('--tts', dest="tts", required=False, type=int, default=0,
                        help="specify the number of codons before TTS will be discarded.. (default: %(default)s AA).")
    parser.add_argument('--stop', dest="stop", action='store_true', required=False, default=False,
                        help="remove the stop codon from table. (default: %(default)s).")
    parser.add_argument('--scale', dest="scale",  choices=['zscore', 'minmax'], required=False, type=str, default='minmax',
                        help="normalize the codon odd ratio. (default: %(default)s).")
    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def rpf_quant_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to quantify RPFs in the CDS region.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('-o', dest="output", required=True, type=str,
                             help="the prefix of output file. (default: prefix + _rpf_quant.txt)")

    parser.add_argument('-f', dest="frame", choices=['0', '1', '2', 'all'], required=False, type=str, default='all',
                        help="set the reading frame for occupancy calculation. (default: %(default)s).")
    parser.add_argument('--tis', dest="tis", required=False, type=int, default=0,
                        help="The number of codons after TIS will be discarded. (default: %(default)s).")
    parser.add_argument('--tts', dest="tts", required=False, type=int, default=0,
                        help="The number of codons before TES will be discarded. (default: %(default)s).")
    parser.add_argument('--utr5', dest="utr5", action='store_true', required=False, default=False,
                        help="quantification of 5'-utr. (default: %(default)s).")
    parser.add_argument('--utr3', dest="utr3", action='store_true', required=False, default=False,
                        help="quantification of 3'-utr. (default: %(default)s).")

    args = parser.parse_args()
    file_check(args.rpf)
    args_print(args)

    return args


def deseq2_args_parser():
    parser = argparse.ArgumentParser(
        description="This script is used to perform differential analysis with riboParser output.")
    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs count file in TXT format.")
    parser.add_argument("-d", dest="design", required=True, type=str, help="input whole design file")
    input_group.add_argument("-o", dest="output", required=True, type=str, help="output the DESeq2 results")

    parser.add_argument("-c", dest="control", required=True, type=str, help="input the control group name")
    parser.add_argument("-t", dest="treat", required=True, type=str, help="input the treat group name")
    parser.add_argument("-m", dest="min", required=False, type=int, default=5,
                        help="specify the minimum reads count. (default: %(default)s).")
    parser.add_argument("-l", dest="logfc", required=False, type=float, default=1,
                        help="specify the logfc threshold for volcano. (default: %(default)s).")
    parser.add_argument("-p", dest="padj", required=False, type=float, default=0.05,
                        help="specify the padj threshold for volcano. (default: %(default)s).")
    args = parser.parse_args()
    file_check(args.rpf)
    file_check(args.design)
    args_print(args)

    return args


def calc_te_args_parser():
    parser = argparse.ArgumentParser(description="Calculate translation efficiency.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-r', dest="ribo", required=True, type=str,
                             help="specify the input Ribo-seq expression table generate by DESeq2.")
    input_group.add_argument('-m', dest="mrna", required=True, type=str, default=None,
                             help="specify the input RNA-seq expression table generate by DESeq2.")
    input_group.add_argument('--rd', dest="ribo_design", required=True, type=str,
                             help="specify the input Ribo-seq expression table generate by DESeq2.")
    input_group.add_argument('--md', dest="mrna_design", required=True, type=str, default=None,
                             help="specify the input RNA-seq expression table generate by DESeq2.")
    input_group.add_argument('-o', dest="output", required=True, type=str, default=None,
                             help="prefix of output filename.")

    # arguments for the RPFs calculation
    parser.add_argument("--min", dest="min", required=False, type=int, default=10,
                        help="specify the minimum reads count. (default: %(default)s).")
    parser.add_argument("-l", dest="logfc", required=False, type=float, default=1,
                        help="specify the logfc threshold for volcano. (default: %(default)s).")
    parser.add_argument("-t", dest="type", choices=['pvalue', 'padj'], required=False, type=str, default='pvalue',
                        help="specify the type of probability value ['pvalue', 'padj']. (default: %(default)s).")
    parser.add_argument("-p", dest="pvalue", required=False, type=float, default=0.05,
                        help="specify the padj threshold for volcano. (default: %(default)s).")
    args = parser.parse_args()

    file_check(args.ribo)
    file_check(args.mrna)
    file_check(args.ribo_design)
    file_check(args.mrna_design)
    args_print(args)

    return args


def ppi_args_parser():
    parser = argparse.ArgumentParser(description="Draw the protein protein interaction network.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-i', dest="input", required=True, type=str, help="specify the input gene list.")
    input_group.add_argument("-x", dest="taxonomy", required=True, type=str,
                             default=None,
                             help="specify the NCBI taxonomy (eg. human: 9606). (default: %(default)s).")
    input_group.add_argument("-p", dest="scopes", required=False,
                             choices=['name', 'entrezgene', 'ensembl', 'symbol', 'uniprot', 'refseq'],
                             default='ensembl',
                             help="specify the input gene type. (default: %(default)s).")
    input_group.add_argument('-o', dest="output", required=True, type=str, default=None,
                             help="prefix of output filename.")

    # arguments for the RPFs calculation
    parser.add_argument('-num', dest="number", required=False, type=str, default=50,
                        help="specify the input number of input gene for network (default: %(default)s).")
    parser.add_argument("-f", dest="flavor", required=False, type=str,
                        choices=['evidence', 'confidence', 'actions'],
                        help="the style of edges in the network: evidence, confidence (default), actions. (default: %(default)s).")
    parser.add_argument("-s", dest="score", required=False, type=int,
                        default=500,
                        help="specify the taxonomy . (default: %(default)s).")
    parser.add_argument("-t", dest="type", required=False, type=str,
                        choices=['functional', 'physical'],
                        default='physical', help="network type (default: %(default)s).")
    parser.add_argument("-n", dest="nodes", required=False, type=int,
                        choices=[0, 1],
                        default=0,
                        help="hides all protein names from the picture (0 or 1) (default: %(default)s).")
    parser.add_argument("-l", dest="labels", required=False, type=int,
                        choices=[0, 1],
                        default=0,
                        help="when provided use submitted names as protein labels in the network image (0 or 1) (default: %(default)s).")
    parser.add_argument("-c", dest="connected", required=False, type=int,
                        choices=[0, 1],
                        default=0,
                        help="hides all proteins that are not connected to any other protein in your network (0 or 1) (default: %(default)s).")
    parser.add_argument("-b", dest="block", required=False, type=int,
                        choices=[0, 1],
                        default=0,
                        help="disables structure pictures inside the bubble (0 or 1) (default: %(default)s).")
    parser.add_argument('-g', dest="graphic", required=False, choices=['highres_image', 'svg'], type=str,
                        default='svg', help="specify the graphic format (default: %(default)s).")
    args = parser.parse_args()

    file_check(args.input)
    args_print(args)

    return args


def serp_peak_args_parser():
    parser = argparse.ArgumentParser(
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=5, width=90),
        description="This script is used to detect the selective translatome.")
    parser.add_argument("-v", "--version", action="version", version="%(prog)s 1.0")
    # arguments for the input files
    input_group = parser.add_argument_group('Input files arguments')
    input_group.add_argument("-r", dest="rpf", required=True, type=str, help="input the RPF coverage file")
    input_group.add_argument("-n", dest="norm", required=False, type=str, default=None,
                             help="specify the file contain RPFs number for normalise. "
                                  "(default: The total RPFs count of input samples is used.)")
    input_group.add_argument("--scale", dest="scale", required=False, type=float, default=1e6,
                             help="specify the scale level for reads normalise (default RPM %(default)s )")
    input_group.add_argument("-a", dest="anno", required=False, type=str,
                             help="specify the gene annotation file in txt format.")
    input_group.add_argument("--ck", dest="control", required=True, type=str,
                             help="specify the name of control samples, separated by commas.")
    input_group.add_argument("--ip", dest="ip", required=True, type=str,
                             help="specify the name of immunoprecipitation samples, separated by commas.")

    # arguments for the data filtering
    data_group = parser.add_argument_group('Data filtering arguments')
    data_group.add_argument("-m", dest="min", required=False, type=int, default=50,
                            help='''specify the minimum number of RPFs coverage for genes to be retained
             (default %(default)s)''')
    data_group.add_argument("--corr", dest="corr", required=False, type=float, default=0.3,
                            help="specify the minimum correlation of replicates (default %(default)s)")

    # arguments for the peaks scanning, gaps and ratio filtering
    peak_group = parser.add_argument_group('Peak scanning arguments')
    peak_group.add_argument("-f", dest="fill", required=False, type=int, default=30,
                            help='''Use the mean value of gene RPF to fill the missing values.
        (Commonly used first 30 AA, default %(default)s codon). Because the
         ribo-seq data is too noisy in single codon level and many locations 
         are not covered by RPF, it is necessary to fill the blank position 
         use the mean value of RPF as the background. Three modes are provided: 
         (1): [30] average of the first 30 AA  is used (or other length); 
         (2): [-1] the average RPFs value of current gene is used; 
         (3): [0] the average RPF of total genes is used.''')
    peak_group.add_argument("-s", dest="size", required=False, type=int, default=3,
                            help="Specifies the window size, the numbers must be odd. "
                                 "(default %(default)s AA)."
                                 "The sequencing data is usually noisy, and need to be smoothed."
                                 "Mean filtering is used here. ")
    peak_group.add_argument("-w", dest="width", required=False, type=int, default=5,
                            help="specify the width of binding peaks (default %(default)s AA)."
                                 "By default, the width of each peak is not less than 5 amino acids.")
    peak_group.add_argument("-e", dest="enrich", required=False, type=float, default=2.0,
                            help="specify the enrich threshold of each peak height (default %(default)s)")
    peak_group.add_argument("-c", dest="collision", required=False, type=float, default=1.5,
                            help="specify the enrich threshold of collision region of binding peak (default %(default)s)")
    peak_group.add_argument("-g", dest="gaps", required=False, type=int, default=1,
                            help='''specify the gap width inside each peak. (default %(default)s AA).
        The ribo-seq profiling is noisy at the single gene level (for example,
        the ribosomal density along a gene may change from a positive number
        to zero and, again, to a positive number). Therefore, gaps in the peak
        are allowed. By default, the gaps there can be no more than 2 consecutive
        amino acids in a peak.''')
    peak_group.add_argument("-p", dest="proportion", required=False, type=float, default=0.2,
                            help='''specify the proportion of gap in each peak (default %(default)s).
        Data with too much noise has low credibility, so it is necessary to
        add a limit to the gap proportion in each peak. By default, the total gaps
        length within a peak cannot exceed 20%%.''')
    peak_group.add_argument("--back", dest="background", required=False, type=int, default=0,
                            help='''use the 5 prime AA to filter the enrichment fold (default %(default)s codon,
        not applicable). Two modes are provided: 
        (1): [30](or other number > 0) Some proteins that bind to the nascent polypeptide 
        chain cannot affect the translation process of the first 30 AA. Theoretically, 
        the RPFs of the unbound area should be smaller than the RPFs of the bound area. 
        (2): [0] However, some proteins only affect the elongation of the ribosome, so 
        the peak may exist in any location.''')
    peak_group.add_argument("--bf", dest="backFold", required=False, action='store_true', default=True,
                            help='''specify the fold of background (default %(default)s).
        Enrichment after background region need greater than before (HSP70 binding model). 
        backFold used to filter the strong binding peak.
        ''')
    peak_group.add_argument("--up", dest="upstream", required=False, type=int, default=10,
                            help='''retrieve the upstream sequence of binding peaks, (default %(default)s codon.)''')
    peak_group.add_argument("--down", dest="downstream", required=False, type=int, default=10,
                            help='''retrieve the downstream sequence of binding peaks, (default %(default)s codon.)''')
    
    # arguments for output figures, ratio, peak annotation
    output_group = parser.add_argument_group('Output files arguments')
    output_group.add_argument("-o", dest="output", required=False, type=str, default='results',
                              help='''prefix of output file name (default %(default)s_ratio.txt).
        This prefix will be used as the output of Peak/Ratio/Matlab Script/Bed files.''')
    output_group.add_argument("--all", dest="all", required=False, action='store_true', default=False,
                              help='''output all the peak region. If this option on, all peak 
        regions include overlapped items in same gene are retained. Instead, 
        only the optimal peak for non-overlapping regions is output. (default %(default)s)''')
    output_group.add_argument("--rpm", dest="rpm", required=False, action='store_true', default=False,
                              help="output the rpm of each peak. (default %(default)s)")
    output_group.add_argument("--ratio", dest="ratio", required=False, action='store_true', default=False,
                              help="output the original ratio. (default %(default)s)")
    # output_group.add_argument(
    #     "--script", dest="script", required=False, action='store_true', default=True,
    #     help="output the MATLAB scripts for each figure. (default %(default)s)"
    # )
    output_group.add_argument("--fig", dest="fig", required=False, action='store_true', default=False,
                              help="draw the demo graph of peak scan results. (default %(default)s)."
                                   "This step may takes a lot of time.")

    args = parser.parse_args()
    file_check(args.rpf)
    file_check(args.anno)
    args_print(args)

    return args


def serp_overlap():
    parser = argparse.ArgumentParser(description="This script is used to check the overlap of mock-IP and IP.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-m', dest="mock", required=True, type=str,
                             help="the name of input mock-IP peak file in TXT format.")
    input_group.add_argument('-f', dest="flag", required=False, type=str,
                             help="the name of input flag-IP peak file in TXT format.")
    input_group.add_argument('--om', dest="out_mock", required=False, type=str,
                             help="the prefix of mock peak output file (prefix + .overlap.txt).")
    input_group.add_argument('--of', dest="out_flag", required=False, type=str,
                             help="the prefix of flag peak output file (prefix + .overlap.txt).")
    args = parser.parse_args()
    file_check(args.mock)
    file_check(args.flag)
    args_print(args)

    return args


def serp_retrieve():
    parser = argparse.ArgumentParser(description="This script is used to retrieve the peak sequence.")
    input_group = parser.add_argument_group('Required arguments')

    input_group.add_argument('-r', dest="rpf", required=True, type=str,
                             help="the name of input RPFs file in TXT format.")
    input_group.add_argument('-p', dest="peak", required=True, type=str,
                             help="the name of input peak file in TXT format.")
    input_group.add_argument('-o', dest="output", required=False, type=str, default='input',
                             help="output sequence file (default: %(default)s.fa).")
    # arguments for the peak region
    parser.add_argument('-u', dest="upstream", required=False, type=int, default=0,
                        help="set the up stream length. (default: %(default)s AA).")
    parser.add_argument('-d', dest="downstream", required=False, type=int, default=0,
                        help="set the down stream length. (default: %(default)s AA).")
    args = parser.parse_args()
    file_check(args.rpf)
    file_check(args.bed)
    args_print(args)

    return args


def serp_retrieve_seq():
    parser = argparse.ArgumentParser(description="This script is used to retrieve the peak sequence.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-p', dest="peak", required=True, type=str,
                             help="the name of input peak file in TXT format.")
    input_group.add_argument('-f', dest="fasta", required=False, type=str, help="the name of fasta sequence.")
    input_group.add_argument('-o', dest="output", required=False, type=str, help="the prefix of output file.")
    # arguments for the peak region
    parser.add_argument('-u', dest="up", required=False, type=int, default=30,
                        help="set the up stream length. (default: %(default)s nt).")
    parser.add_argument('-d', dest="down", required=False, type=int, default=30,
                        help="set the down stream length. (default: %(default)s nt).")

    args = parser.parse_args()
    file_check(args.peak)
    file_check(args.fasta)
    args_print(args)

    return args


def serp_properties():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="This script is used to evaluate the different properties of sequence.",
                                     epilog=textwrap.dedent('''
        Properties contain the following patrs:
        Part1:
            CAI  - codon adaptation index
            RSCU - Relative Synonymous Codon Usage
            Obsi - Observed number of occurrences of codon (per 1000 codon)
        Part2:
            gravy             - according to Kyte and Doolittle
            structure         - fraction of helix, turn and sheet
            flexibility       - according to Vihinen, 1994
            instability       - according to Guruprasad et al 1990
            isoelectric_point - use module IsoelectricPoint
        '''))

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-f', dest="fasta", required=True, type=str,
                             help="the name of input RNA sequence in fasta format.")
    input_group.add_argument('-o', dest="output", required=True, type=str, help="the prefix of output file,.")
    args = parser.parse_args()
    file_check(args.fasta)
    args_print(args)

    return args



#!/usr/bin/env python
# @Project : riboparser
# @Script  : Retrieve.py

import pandas as pd
from . import RPFs


class Retrieve(object):

    def __init__(self, args):
        # input and output
        self.rpf = args.rpf
        self.out = args.output

        self.list = args.list
        self.gene_rpf = None

        # parameters for rpfs filter
        self.rpf_num = args.min
        self.norm = args.normal

        # format rpf table
        self.split = args.split
        self.format = args.format
        
        # rpf table
        self.sample_name = None
        self.sample_num = None
        self.total_rpf_num = None
        self.raw_rpf = None

    def import_gene_list(self):
        '''
        @Message  : import the gene list.
        @Input    : self.list --> gene list in one column
        @Return   : self.gene --> gene list, [YDL246C, YDL243C ...]
        '''
        if self.list is None:
            self.gene = None
        else:
            gene_csv = pd.read_csv(self.list, header=None, names=None)
            self.gene = gene_csv.iloc[:, 0].to_list()

    def retrieve_rpf(self):
        '''
        @Message  : import the rpf table.
        @Input    : self.rpf --> rpf table
                    self.site --> [E, P, A] site in ribosome
                    self.frame --> [0, 1, 2] frame in [E, P, A] site
                    self.gene --> gene list
                    self.rpf_num --> minimum rpf value
                    norm --> normalize the rpf value to RPM
        @Return   : self.raw_rpf --> rpf table contain all rpf value
                    self.sample_name --> samples name
                    self.sample_num --> groups sample number
                    self.merged_rpf --> merged rpf table
                    self.total_rpf_num --> total rpf number of each sample
        '''
        
        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf,
                                      sample_num=None,
                                      sample_name=None,
                                      sites='P',
                                      frame='all',
                                      gene=self.list,
                                      rpf_num=self.rpf_num)

        self.raw_rpf = rpf_results[0].to_pandas()
        self.sample_name = rpf_results[1]
        self.sample_num = rpf_results[2]
        # self.merged_rpf = rpf_results[3]
        self.total_rpf_num = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5]
        self.high_gene = rpf_results[6]

        self.high_rpf = self.raw_rpf.loc[self.raw_rpf['name'].isin(self.high_gene), ]

        del rpf_results

    def rpf_to_rpm(self):
        '''
        @Message  : convert rpf to rpm value.
        @Input    : high_rpf --> rpf table in pandas dataframe
                    norm --> default [False]
        @Return   : high_rpf --> rpm table
        '''
        
        if self.norm:
            total_rpf_num = [x for x in self.total_rpf_num for _ in range(3)]
            #　print the class of data type of eaech column
            now_cols = self.high_rpf.columns[6:].to_list()
            self.high_rpf[now_cols] = self.high_rpf[now_cols].apply(pd.to_numeric, downcast='float')
            self.high_rpf.iloc[:, 6:] = (self.high_rpf.iloc[:, 6:] * 1000000) / total_rpf_num
        else:
            pass

    def melt_rpf_table(self):
        '''
        @Message  : melt the data into a format that is convenient for drawing.
        @Input    : self.high_rpf --> specific genes rpf table 
        @Return   : self.melt_rpf --> melt the three frame data to long data format

        from
            name ... wt_f0 wt_f1 wt_f2
            ORF1 ... 10 3   2
            ORF1 ... 7 2   2
        to
            name ... wt
            ORF1 ... 10
            ORF1 ... 3
            ORF1 ... 2
            ...
        '''
        
        if self.format:
            ids_vars = self.high_rpf.columns[0:6].to_list()
            sample_title = [x + y for x in self.sample_name for y in ['_f0', '_f1', '_f2']]
            self.gene_rpf = pd.melt(self.high_rpf, id_vars=ids_vars, value_vars=sample_title, var_name='sample', value_name='rpf')
            
            self.gene_rpf['frame'] = self.gene_rpf['sample'].str[-1]
            self.gene_rpf['sample'] = self.gene_rpf['sample'].str[:-3]

            self.gene_rpf = self.gene_rpf.sort_values(by=['sample', 'name', 'now_nt', 'frame']).reset_index(drop=True)
            self.gene_rpf = self.gene_rpf.pivot(index=ids_vars + ['frame'], columns='sample', values='rpf').reset_index()
            self.gene_rpf['now_nt'] = self.gene_rpf['now_nt'].astype(int) + self.gene_rpf['frame'].astype(int)
            self.gene_rpf['codon'] = self.gene_rpf.apply(lambda row: row['codon'][int(row['frame'])], axis=1)
            
        else:
            self.gene_rpf = self.high_rpf

    def output_rpf_table(self):
        '''
        @Message  : split merged rpf table to each file.
        @Input    : self.melt_rpf --> description
        @Return   : output --> output rpf table in txt format
        '''
        
        if self.split:
            gene_group = self.gene_rpf.groupby('name')
            for gene, expr in gene_group:
                filename = gene
                expr.to_csv(filename + "_retrieve.txt", sep="\t", index=False)
        else:
            self.gene_rpf.to_csv(self.out + "_retrieve.txt", sep="\t", index=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_CST.py


from .ribo import ArgsParser
from .ribo import CST


def main():
    ArgsParser.now_time()
    print('\nCalculate the codon decoding time.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.rpf_cst_args_parser()

    cst = CST.CodonSelectiveTime(args)

    print('\nStep2: Import the Ribo and mRNA density file.', flush=True)
    cst.import_density()

    # print('\nStep3: Combine the Ribo and mRNA density file.', flush=True)
    # cst.merge_rpf_rna()

    print('\nStep3: Calculate the codon selection time.', flush=True)
    cst.calc_cst()

    print('\nStep4: Output the codon selection time.', flush=True)
    cst.format_cst_results()
    cst.output_cst()

    print('\nStep5: Draw the codon decoding time plot.', flush=True)
    # cst.draw_cst_line()
    cst.draw_cst_corr()
    cst.draw_cst_heat()

    print('All done.\n', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()
